import pytest

import numpy as np
from yabf import chi2, run_map


def create_mock_data(fiducial_fg_logpoly):
    spec = fiducial_fg_logpoly()
    assert len(spec["LogPoly_spectrum"]) == 100
    return spec["LogPoly_spectrum"]


def test_retrieve_params(fiducial_fg_logpoly):
    spec = create_mock_data(fiducial_fg_logpoly)
    lk = chi2.MultiComponentChi2(
        kind="spectrum", components=[fiducial_fg_logpoly], data=spec
    )
    a = run_map(lk)
    assert a.success
    assert np.allclose(a.x, [2, -2.5, 50])
    assert len(a.x) == 3


def test_damped_oscillations(fiducial_dampedoscillations):
    spec = fiducial_dampedoscillations()
    print(spec)
    assert np.allclose(spec["DampedOscillations_spectrum"], 0)
    assert len(spec["DampedOscillations_spectrum"]) == 100



import pytest

import matplotlib as mpl
import numpy as np
from astropy import units as u
from edges_cal.modelling import Polynomial, UnitTransform
from edges_cal.simulate import simulate_qant_from_calobs
from helpers import get_tns_model, sim_antenna_q
from scipy import stats
from yabf import ParamVec, run_map

from edges_estimate.eor_models import AbsorptionProfile
from edges_estimate.likelihoods import DataCalibrationLikelihood


@pytest.fixture(scope="function")
def calobs_freq(calobs):
    return calobs.freq.freq.to_value("MHz")


@pytest.fixture(scope="function")
def calobs_freq_smoothed8(calobs):
    return calobs.freq.freq[::8].to_value("MHz")


@pytest.fixture(scope="function")
def sky_freq():
    return np.linspace(58, 93, 120)


def get_likelihood(
    labcal,
    calobs,
    qvar_ant,
    fg,
    eor,
    cal_noise,
    simulate=True,
    ideal_tns=False,
    loss=1,
    bm_corr=1,
    seed=1234,
):
    q = sim_antenna_q(
        labcal, calobs, fg, eor, ideal_tns=ideal_tns, loss=loss, bm_corr=bm_corr
    )

    if isinstance(qvar_ant, (int, float)):
        qvar_ant = qvar_ant * np.ones(len(eor.freqs))

    if seed:
        np.random.seed(seed)

    q = q + np.random.normal(scale=qvar_ant)

    tns_model, tns_params = get_tns_model(calobs, ideal=ideal_tns)

    return DataCalibrationLikelihood.from_labcal(
        labcal,
        calobs,
        q_ant=q,
        qvar_ant=qvar_ant,
        fg_model=fg,
        eor_components=(eor,),
        as_sim="all" if simulate else (),
        t_ns_params=tns_params,
        cal_noise=cal_noise,
        field_freq=eor.freqs * u.MHz,
        loss=loss,
        bm_corr=bm_corr,
        loss_temp=296 * (1 - loss),
    )


def get_eor(freqs):
    return AbsorptionProfile(
        freqs=freqs,
        params={
            "A": {
                "fiducial": 0.5,
                "min": 0,
                "max": 1.5,
                "ref": stats.norm(0.5, scale=0.01),
            },
            "w": {
                "fiducial": 15,
                "min": 5,
                "max": 25,
                "ref": stats.norm(15, scale=0.1),
            },
            "tau": {
                "fiducial": 5,
                "min": 0,
                "max": 20,
                "ref": stats.norm(5, scale=0.1),
            },
            "nu0": {
                "fiducial": 78,
                "min": 60,
                "max": 90,
                "ref": stats.norm(78, scale=0.1),
            },
        },
    )


def view_results(
    lk, res_data, calobs, eor, plt, sim_tns=True, label=None, fig=None, ax=None, c=0
):
    """Simple function to create a plot of input vs expected TNS and T21."""
    eorspec = lk.partial_linear_model.get_ctx(params=res_data.x)

    if fig is None:
        plot_input = True
        fig, ax = plt.subplots(2, 2, figsize=(15, 7), sharex=True)
    else:
        plot_input = False

    color = f"C{c}"
    nu = calobs.freq.freq

    tns_model, _ = get_tns_model(calobs, ideal=sim_tns)
    tns_model = tns_model(nu)

    if plot_input:
        ax[0, 0].plot(nu, tns_model, label="Input", color="k")

    ax[0, 0].plot(
        nu,
        eorspec["tns"],
        label="Estimated" + (" " + label if label else ""),
        color=color,
    )

    ax[1, 0].plot(
        nu,
        eorspec["tns"] - tns_model,
        label=r"$\Delta T_{\rm NS}$" if plot_input else None,
        color=color,
    )

    ax[0, 0].set_title(r"$T_{\rm NS}$")
    ax[0, 0].set_ylabel("Temperature [K]")

    nu = eor.freqs

    if plot_input:
        ax[0, 1].plot(nu, eor()["eor_spectrum"], color="k")

    ax[0, 1].plot(nu, eorspec["eor_spectrum"])
    ax[0, 1].set_title(r"$T_{21}$")
    delta = eorspec["eor_spectrum"] - eor()["eor_spectrum"]
    ax[1, 1].plot(
        nu,
        delta,
        color=color,
        label=rf"Max $\Delta = {np.max(np.abs(delta))*1000:1.2e}$mK",
    )
    ax[1, 0].set_ylabel("Difference [K]")

    ax[1, 0].set_xlabel("Frequency")
    ax[1, 1].set_xlabel("Frequency")

    ax[0, 0].legend()
    ax[1, 0].legend()
    ax[1, 1].legend()

    return fig, ax


# Define a loss function gotten from fitting to Alan's loss model (not necessary that
# its accurate, just realistic to make the test useful)
data_like_loss = Polynomial(
    parameters=[
        9.94009502e-01,
        7.17401803e-04,
        4.50954613e-03,
        -5.83728183e-03,
        -2.10009238e-03,
        3.81924327e-03,
        -1.58099823e-03,
        4.37297391e-04,
    ],
    n_terms=8,
    transform=UnitTransform(range=[50, 100]),
)

data_like_bmcorr = Polynomial(
    parameters=[
        1.00000433,
        -0.00790527,
        -0.00566472,
        -0.00278198,
        0.0016465,
        -0.00294487,
        -0.00508308,
        0.00379791,
        0.00293712,
        -0.00173159,
    ],
    n_terms=10,
    transform=UnitTransform(range=[50, 100]),
)


def unity_loss(x):
    return 1


@pytest.mark.parametrize(
    "lc,cl,qvar_ant,cal_noise,simulate,atol,fsky,loss,bm_corr",
    [
        (
            "labcal",
            "calobs",
            0.0,
            0.0,
            True,
            0.01,
            "calobs_freq",
            unity_loss,
            unity_loss,
        ),  # No noise
        (
            "labcal",
            "calobs",
            1e-10,
            1e-10,
            True,
            0.01,
            "calobs_freq",
            unity_loss,
            unity_loss,
        ),  # Small constant noise
        (
            "labcal",
            "calobs",
            1e-10,
            "data",
            True,
            0.01,
            "calobs_freq",
            unity_loss,
            unity_loss,
        ),  # Realistic non-constant noise on smooth cal solutions
        (
            "labcal12",
            "calobs12",
            1e-10,
            "data",
            False,
            0.05,
            "calobs_freq",
            unity_loss,
            unity_loss,
        ),  # Actual cal data
        (
            "labcal",
            "calobs",
            1e-10,
            "data",
            True,
            0.01,
            "calobs_freq_smoothed8",
            unity_loss,
            unity_loss,
        ),  # Realistic non-constant noise on smooth cal solutions with fewer sky freqs
        (
            "labcal",
            "calobs",
            1e-12,
            "data",
            True,
            0.01,
            "sky_freq",
            unity_loss,
            unity_loss,
        ),  # Realistic non-constant noise on smooth cal solutions with different freq range
        (
            "labcal",
            "calobs",
            1e-10,
            "data",
            True,
            0.01,
            "calobs_freq",
            data_like_loss,
            unity_loss,
        ),  # Realistic non-constant noise on smooth cal solutions WITH LOSS
        (
            "labcal",
            "calobs",
            1e-10,
            "data",
            True,
            0.01,
            "calobs_freq",
            data_like_loss,
            data_like_bmcorr,
        ),  # Realistic non-constant noise on smooth cal solutions WITH LOSS AND BMCORR
    ],
)
def test_cal_data_likelihood(
    lc,
    cl,
    fiducial_fg,
    qvar_ant,
    cal_noise,
    simulate,
    atol,
    fsky,
    loss,
    bm_corr,
    request,
    plt,
):
    fsky = request.getfixturevalue(fsky)
    labcal = request.getfixturevalue(lc)
    calobs = request.getfixturevalue(cl)
    eor = get_eor(fsky)

    lk = get_likelihood(
        labcal,
        calobs,
        qvar_ant=qvar_ant,
        fg=fiducial_fg,
        eor=eor,
        cal_noise=cal_noise,
        simulate=simulate,
        loss=loss(fsky),
        bm_corr=bm_corr(fsky),
        ideal_tns=False,
    )

    res = run_map(lk.partial_linear_model)
    eorspec = lk.partial_linear_model.get_ctx(params=res.x)

    tns_model, _ = get_tns_model(calobs, ideal=False)
    tns_model = tns_model(labcal.calobs.freq.freq.to_value("MHz"))

    if plt == mpl.pyplot:
        view_results(lk, res, calobs, eor, plt, sim_tns=simulate)

    np.testing.assert_allclose(tns_model, eorspec["tns"], atol=0, rtol=1e-2)
    np.testing.assert_allclose(
        eor()["eor_spectrum"], eorspec["eor_spectrum"], atol=atol, rtol=0
    )



import pytest

import numpy as np

from edges_estimate import plots as p


@pytest.fixture(scope="session")
def models():
    return {key: {"model": np.random.normal(size=(20, 50))} for key in ["one", "two"]}


@pytest.fixture(scope="session")
def freqs():
    return {key: np.random.normal(size=50) for key in ["one", "two"]}


@pytest.fixture(scope="session")
def temp():
    return {key: np.random.normal(size=50) for key in ["one", "two"]}


def test_make_residual_plot_shaded(models, freqs, temp):
    p.make_residual_plot_shaded(models, freqs, temp)



import numpy as np
from getdist import loadMCSamples
from matplotlib import pyplot as plt
from tqdm import tqdm
from yabf import LikelihoodContainer, load_likelihood_from_yaml


def get_evidence(root: str):
    """Read the Bayesian evidence of the Polychord run

    Parameters
    ----------
    root
        The root file path of the run without the extension
    """

    with open(root + ".stats") as f:
        for line in f:
            if line.startswith("log(Z)"):
                evidence = float(line.split("=")[1].lstrip().split(" ")[0])
                break
    return evidence


def get_models_from_mcsamples(
    mcsamples, lk_names, lk=None, extras=None, n=None, progress=True
):
    if type(mcsamples) == str:
        mcsamples = loadMCSamples(mcsamples)

    if n is not None:
        samples = mcsamples.samples[
            np.random.choice(mcsamples.samples.shape[0], size=n)
        ]
    else:
        samples = mcsamples.samples

    out = {}
    if lk is None:
        lk = load_likelihood_from_yaml(mcsamples.rootname + "_config.yml")

    top_level = True
    if isinstance(lk, LikelihoodContainer):
        top_level = False

    if top_level and len(lk_names) > 1:
        raise ValueError("you have specified more lk_names than actually exist!")

    for params in tqdm(samples, disable=not progress):
        ctx = lk.get_ctx(params=params)
        model = lk.reduce_model(ctx=ctx, params=params)

        if top_level:
            ctx = {lk_names[0]: ctx}
            model = {lk_names[0]: model}

        for lk_name in lk_names:

            if lk_name not in out:
                out[lk_name] = {}

            if "model" not in out[lk_name]:
                out[lk_name]["model"] = []

            out[lk_name]["model"].append(model[lk_name])

            for extra in extras or []:
                if extra in ctx[lk_name]:
                    out[lk_name][extra] = ctx[lk_name][extra]

    return out


def make_residual_plot_shaded(models, freqs=None, temps=None, labels=None, color=None):
    labels = labels or {}
    for i, (key, freq) in enumerate(freqs.items()):
        model = models[key]["model"]
        temp = temps[key]
        label = labels.get(key, key)

        q = np.quantile(model, q=(0.04, 0.16, 0.5, 0.84, 0.96), axis=0)

        plt.plot(freq, temp - q[2], color=color or f"C{i}", label=label)
        plt.fill_between(
            freq, temp - q[1], temp - q[3], color=color or f"C{i}", alpha=0.6
        )
        plt.fill_between(
            freq, temp - q[0], temp - q[4], color=color or f"C{i}", alpha=0.4
        )



import pytest

import hickle
import numpy as np
from edges_analysis.analysis.calibrate import LabCalibration
from edges_cal import Calibrator
from edges_cal.modelling import LinLog
from pathlib import Path
from scipy import stats

from edges_estimate.eor_models import AbsorptionProfile
from edges_estimate.foregrounds import DampedOscillations, LogPoly


@pytest.fixture(scope="session")
def data_path() -> Path:
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def calobs(data_path) -> Calibrator:
    return hickle.load(data_path / "test_calfile.h5")


@pytest.fixture(scope="session")
def labcal(calobs, data_path) -> Calibrator:
    return LabCalibration.from_s11_files(
        calobs=calobs, s11_files=sorted(data_path.glob("*.s1p"))
    )


@pytest.fixture(scope="session")
def calobs12(calobs) -> Calibrator:
    """Calobs with 12 c/w terms."""
    return calobs.clone(cterms=12, wterms=12)


@pytest.fixture(scope="session")
def labcal12(calobs12, data_path) -> Calibrator:
    return LabCalibration.from_s11_files(
        calobs=calobs12, s11_files=sorted(data_path.glob("*.s1p"))
    )


@pytest.fixture(scope="session")
def fiducial_eor(calobs):
    return AbsorptionProfile(
        freqs=calobs.freq.freq,
        params={
            "A": {
                "fiducial": 0.5,
                "min": 0,
                "max": 1.5,
                "ref": stats.norm(0.5, scale=0.01),
            },
            "w": {
                "fiducial": 15,
                "min": 5,
                "max": 25,
                "ref": stats.norm(15, scale=0.1),
            },
            "tau": {
                "fiducial": 5,
                "min": 0,
                "max": 20,
                "ref": stats.norm(5, scale=0.1),
            },
            "nu0": {
                "fiducial": 78,
                "min": 60,
                "max": 90,
                "ref": stats.norm(78, scale=0.1),
            },
        },
    )


@pytest.fixture(scope="function")
def fiducial_fg():
    return LinLog(n_terms=5, parameters=[2000, 10, -10, 5, -5])


@pytest.fixture(scope="function")
def fiducial_fg_logpoly():
    return LogPoly(
        freqs=np.linspace(50, 100, 100),
        poly_order=2,
        params={
            "p0": {"fiducial": 2, "min": -5, "max": 5},
            "p1": {"fiducial": -2.5, "min": -3, "max": -2},
            "p2": {"fiducial": 50, "min": -100, "max": 100},
        },
    )


@pytest.fixture(scope="function")
def fiducial_dampedoscillations():
    return DampedOscillations(
        freqs=np.linspace(50, 100, 100),
        params={
            "amp_sin": {"fiducial": 0, "min": -5, "max": 5},
            "amp_cos": {"fiducial": 0, "min": -3, "max": -2},
            "P": {"fiducial": 15, "min": -100, "max": 100},
            "b": {"fiducial": 1, "min": -100, "max": 100},
        },
    )



import click
import numpy as np
import os
import sys
from edges_cal import CalibrationObservation
from getdist import plots
from pathlib import Path
from yabf import samplers
from yabf.core import yaml

from edges_estimate.calibration import CalibratorQ
from edges_estimate.likelihoods import CalibrationChi2


@click.command()
@click.argument("path", type=click.Path(exists=True, file_okay=False))
@click.option("-l", "--freq-low", type=float, default=50)
@click.option("-h", "--freq-high", type=float, default=100)
@click.option("-c", "--cterms", type=int, default=10)
@click.option("-w", "--wterms", type=int, default=12)
@click.option("-m", "--use-var-model", type=bool, default=False)
@click.option("-b/-B", "--bounds/--no-bounds", default=True)
def main(path, freq_low, freq_high, cterms, wterms, use_var_model, bounds):

    direc = Path(__file__).parent.absolute()
    path = Path(path)

    if "2015_09" in str(path):
        resistance_f = 49.98
        resistance_m = 50.12
    else:
        resistance_f = 49.999
        resistance_m = 50.1501

    calobs = CalibrationObservation(
        path=path,
        f_low=freq_low,
        f_high=freq_high,
        run_num=None,
        cterms=cterms,
        wterms=wterms,
        resistance_f=resistance_f,
        load_kwargs={
            "cache_dir": "/data4/smurray/edges_cal_cache",
            "ignore_times_percent": 5,
        },
        s11_kwargs={"resistance": resistance_m},
        include_previous=False,
        compile_from_def=False,
    )

    fig = calobs.plot_coefficients()
    fig.savefig("coefficients.pdf")
    fig = calobs.plot_calibrated_temps()
    fig.savefig("calibrated_temps.pdf")

    # Create a CalibratorQ component that has all the active parameters, but
    # is very basic.
    cq = CalibratorQ(
        calobs=calobs,
        params={
            **{
                f"C1_{i}": {"fiducial": calobs.C1_poly.coefficients[-(i + 1)]}
                for i in range(calobs.cterms)
            },
            **{
                f"C2_{i}": {"fiducial": calobs.C2_poly.coefficients[-(i + 1)]}
                for i in range(calobs.cterms)
            },
            **{
                f"Tunc_{i}": {"fiducial": calobs.Tunc_poly.coefficients[-(i + 1)]}
                for i in range(calobs.wterms)
            },
            **{
                f"Tcos_{i}": {"fiducial": calobs.Tcos_poly.coefficients[-(i + 1)]}
                for i in range(calobs.wterms)
            },
            **{
                f"Tsin_{i}": {"fiducial": calobs.Tsin_poly.coefficients[-(i + 1)]}
                for i in range(calobs.wterms)
            },
        },
    )

    if not use_var_model:
        # Create a likelihood that has data that is Qp (we need to choose Tns and Tload here)
        lk = CalibrationChi2(
            components=(cq,),
            use_model_sigma=False,
            sigma={
                k: np.sqrt(spec.spectrum.variance_Q)
                for k, spec in calobs._loads.items()
            },
            data={k: spec.spectrum.averaged_Q for k, spec in calobs._loads.items()},
        )

    file_name = (
        direc
        / f"{path.name}_l{freq_low}MHz_h{freq_high}MHz_c{cterms}_w{wterms}{'_use_model' if use_var_model else ''}{'bounds' if bounds else '_no_bounds'}"
    )

    if not file_name.exists():
        os.mkdir(file_name)

    # Write out necessary data files
    np.savez(file_name / "data.npz", **lk.data)
    np.savez(file_name / "sigma.npz", **lk.sigma)

    prms = {}
    for kind in ["C1", "C2", "Tunc", "Tcos", "Tsin"]:
        if bounds:
            prms[kind] = "\n          ".join(
                f"{kind}_{i}:\n            min: {coeff - 20*np.abs(coeff)}\n            max: {coeff + 20*np.abs(coeff)}\n            fiducial: {coeff}"
                for i, coeff in enumerate(
                    getattr(calobs, f"{kind}_poly").coefficients[::-1]
                )
            )
        else:
            prms[kind] = "\n          ".join(
                f"{kind}_{i}:\n            fiducial: {coeff}"
                for i, coeff in enumerate(
                    getattr(calobs, f"{kind}_poly").coefficients[::-1]
                )
            )

    pwd = os.getcwd()

    config = f"""
name: {file_name.name}
external_modules:
- edges_estimate
likelihoods:
  calibration:
    class: CalibrationChi2
    components:
      calibrator:
        class: CalibratorQ
        kwargs:
          path: {path}
          calobs_args:
            f_low: {freq_low}
            f_high: {calobs.freq.max}
            cterms: {cterms}
            wterms: {wterms}
            resistance_f: {calobs.lna.resistance}
            s11_kwargs:
              resistance: {calobs.open.reflections.resistance}
            load_kwargs:
              ignore_times_percent: {calobs.open.spectrum.ignore_times_percent}
              cache_dir: {calobs.open.spectrum.cache_dir}
            load_s11s:
              open:
                n_terms: 151
            compile_from_def: false
        params:
          {prms['C1']}
          {prms['C2']}
          {prms['Tunc']}
          {prms['Tcos']}
          {prms['Tsin']}
    kwargs:
      use_model_sigma: {use_var_model}
      sigma: {os.path.join(pwd, file_name, 'sigma.npz')}
    data: {os.path.join(pwd, file_name, 'data.npz')}
"""

    with open(file_name / "config.yml", "w") as fl:
        fl.write(config)

    print(f"Wrote config to {file_name / 'config.yml'}")


if __name__ == "__main__":
    main()



"""
Models of the foregrounds
"""
import attr
import numpy as np
from cached_property import cached_property
from yabf import Component, Parameter


@attr.s(frozen=True)
class Foreground(Component):
    """Base class for all foreground models, don't use this directly!"""

    freqs: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))
    nuc = attr.ib(75.0, kw_only=True, converter=float)

    @cached_property
    def provides(self):
        return [f"{self.name}_spectrum"]

    @cached_property
    def f(self):
        return self.freqs / self.nuc

    def calculate(self, ctx=None, **params):
        return self.model(**params)

    def model(self, **params):
        pass


@attr.s(frozen=True)
class Tcmb(Component):
    T = attr.ib(2.7255, kw_only=True, converter=float)

    @cached_property
    def provides(self):
        return [f"{self.name}_scalar"]

    def calculate(self, ctx=None, **params):
        return self.T


class _PhysicalBase(Foreground):
    base_parameters = [
        Parameter("b0", 1750, min=0, max=1e5, latex=r"b_0 [K]"),
        Parameter("b1", 0, latex=r"b_1 [K]"),
        Parameter("b2", 0, latex=r"b_2 [K]"),
        Parameter("b3", 0, latex=r"b_3 [K]"),
        Parameter("ion_spec_index", -2, min=-3, max=-1, latex=r"\alpha_{\rm ion}"),
    ]

    def model(self, **p):
        b = [p[f"b{i}"] for i in range(4)]
        alpha = p["ion_spec_index"]

        x = np.exp(-b[3] * self.f**alpha)
        return b[0] * self.f ** (b[1] + b[2] * np.log(self.f) - 2.5) * x, x


class PhysicalHills(_PhysicalBase):
    """
    Eq 6. from Hills et al.
    """

    base_parameters = _PhysicalBase.base_parameters + [
        Parameter("Te", 1000, min=0, max=5000, latex=r"T_e [K]"),
    ]

    def model(self, Te, **p):
        first_term, x = super().model(**p)
        return first_term + Te * (1 - x)


class PhysicalSmallIonDepth(_PhysicalBase):
    """
    Eq. 7 from Hills et al.
    """

    base_parameters = _PhysicalBase.base_parameters + [
        Parameter("b4", 0, latex=r"b_4 [K]")
    ]

    def model(self, **p):
        first_term, x = super().model(p)
        b4 = p["b4"]
        return first_term + b4 / self.f**2

    # Possible derived quantities
    def Te(self, ctx, **params):
        """Approximate value of Te in the small-ion-depth limit"""
        return params["b4"] / params["b3"]


class PhysicalLin(Foreground):
    """
    Eq. 8 from Hills et al.
    """

    base_parameters = [Parameter("p0", fiducial=1750, latex=r"p_0")] + [
        Parameter(f"p{i}", fiducial=0, latex=rf"p_{i}") for i in range(1, 5)
    ]

    def model(self, **p):
        p = [p[f"p{i}"] for i in range(5)]

        return (
            self.f**-2.5 * (p[0] + np.log(self.f) * (p[1] + p[2] * np.log(self.f)))
            + p[3] * self.f**-4.5
            + p[4] * self.f**-2
        )

    # Possible derived parameters
    def b0(self, ctx, **p):
        """The corresponding b0 from PhysicalHills"""
        return p["p0"]

    def b1(self, ctx, **p):
        """The corresponding b1 from PhysicalHills"""
        return p["p1"] / p["p0"]

    def b2(self, ctx, **p):
        """The corresponding b2 from PhysicalHills"""
        return p["p2"] / p["p0"] - self.b1(ctx, **p) ** 2 / 2

    def b3(self, ctx, **p):
        """The corresponding b3 from PhysicalHills"""
        return -p["p3"] / p["p0"]

    def b4(self, ctx, **p):
        """The corresponding b4 from PhysicalHills"""
        return p["p4"]


@attr.s
class IonContrib(Foreground):
    """
    Absorption and emission due to the ionosphere
    """

    base_parameters = [Parameter("absorption", fiducial=0, latex=r"\tau")] + [
        Parameter("emissivity", fiducial=0, latex=r"T_{elec}")
    ]

    def model(self, **p):
        return p["absorption"] * self.f**-4.5 + p["emissivity"] * self.f**-2


@attr.s
class LinLog(Foreground):
    """
    LinLog model from Memo #122

    .. note:: the model there is slightly ambiguous. Actually taking the Taylor Expansion
              of the ExpLog model makes it clear that a_1 (here p_1) should be
              equivalently zero. See Memo #122 for details.

              We leave that parameter open, but suggest not letting it vary, and leaving
              it as zero.

    Parameters
    ----------
    poly_order : int
        The maximum polynomial order will be `poly_order - 1`. There are `poly_order + 1`
        total parameters, including `beta` and `p1` (which should usually be set to zero).
    """

    poly_order = attr.ib(5, converter=int, kw_only=True)
    use_p1 = attr.ib(False, converter=bool)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        if not self.use_p1 and "p1" in self.child_active_params:
            raise ValueError(
                "You are attempting to fit p1, but it won't affect anything!"
            )

    @cached_property
    def base_parameters(self):
        p = [
            Parameter("beta", -2.5, min=-5, max=0, latex=r"\beta"),
            Parameter("p0", 1750, latex=r"p_0"),
            Parameter("p1", 0, latex=r"p_1"),
        ]

        assert self.poly_order >= 1, "poly_order must be >= 1"

        # First create the parameters.
        for i in range(2, self.poly_order):
            p.append(Parameter(f"p{i}", 0, latex=rf"p_{i}"))
        return tuple(p)

    @cached_property
    def logf(self):
        return np.log(self.f)

    @cached_property
    def basis(self):
        return np.array([self.logf**i for i in range(self.poly_order)])

    def model(self, **p):
        pp = [p[f"p{i}"] for i in range(self.poly_order)]
        if not self.use_p1:
            pp[1] = 0

        terms = [pp[i] * self.basis[i] for i in range(self.poly_order)]
        return self.f ** p["beta"] * np.sum(terms, axis=0)


@attr.s
class Sinusoid(Foreground):
    base_parameters = [
        Parameter("amp", 0, min=0, max=1, latex=r"A_{\rm sin}"),
        Parameter("lambda", 10, min=1, max=30, latex=r"\lambda_{\rm sin}"),
        Parameter("phase", 0, min=-np.pi, max=np.pi, latex=r"\phi_{\rm sin}"),
    ]

    def model(self, **p):
        return p["amp"] * np.sin(2 * np.pi * self.freqs / p["lambda"] + p["phase"])


@attr.s
class DampedOscillations(Foreground):
    base_parameters = [
        Parameter("amp_sin", 10e-10, min=-10, max=1000, latex=r"A_{\rm sin}"),
        Parameter("amp_cos", 10e-10, min=-10, max=1000, latex=r"A_{\rm cos}"),
        Parameter("P", 10, min=1, max=np.inf, latex=r"P_{\rm MHz}"),
        Parameter("b", 0, min=-10, max=10, latex=r"b"),
    ]

    def model(self, **p):
        phase = 2 * np.pi * self.freqs / p["P"]
        return (self.f) ** p["b"] * (
            p["amp_sin"] * np.sin(phase) + p["amp_cos"] * np.cos(phase)
        )


@attr.s
class DampedSinusoid(Component):
    freqs: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))
    provides = ("sin_spectrum",)

    base_parameters = [
        Parameter("amp", 0, min=0, max=1, latex=r"A_{\rm sin}"),
        Parameter("lambda", 10, min=1, max=30, latex=r"\lambda_{\rm sin}"),
        Parameter("phase", 0, min=-np.pi, max=np.pi, latex=r"\phi_{\rm sin}"),
    ]

    def calculate(self, ctx=None, **p):
        models = np.array([v for k, v in ctx.items() if k.endswith("spectrum")])
        amp = np.sum(models, axis=0)
        amp *= p["amp"]
        return amp * np.sin(2 * np.pi * self.freqs / p["lambda"] + p["phase"])


class LinPoly(LinLog):
    """Eq. 10 from Hills et al.

    .. note:: The polynomial terms are offset by a prior assumption for beta: -2.5.

    The equation is

    .. math :: T(nu) = (nu/nuc)**-beta * Sum[p_i (nu/nuc)**i]
    """

    def model(self, **p):
        terms = []
        for key, val in p.items():
            if key == "beta":
                continue
            i = int(key[1:])
            terms.append(val * self.f**i)

        return np.sum(terms, axis=0) * self.f ** p["beta"]


@attr.s
class Bias(Component):
    x: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))
    centre = attr.ib(1, converter=float, kw_only=True)

    poly_order = attr.ib(1, converter=int, kw_only=True)
    kind = attr.ib("spectrum", kw_only=True)
    log = attr.ib(False, kw_only=True)
    additive = attr.ib(False, kw_only=True, converter=bool)

    @cached_property
    def base_parameters(self):
        p = [Parameter("b0", 1, min=-np.inf if self.additive else 0, latex=r"b_0")]

        assert self.poly_order >= 1, "poly_order must be >= 1"

        # First create the parameters.
        for i in range(1, self.poly_order):
            p.append(Parameter(f"b{i}", 0, latex=rf"b_{i}"))
        return tuple(p)

    def evaluate_poly(self, **params):
        x = self.x / self.centre
        if self.log:
            x = np.log(x)

        res = 0
        for i in range(self.poly_order):
            p = params[f"b{i}"]
            res += p * x**i

        return res

    def calculate(self, ctx, **params):
        bias = self.evaluate_poly(**params)

        for key, val in ctx.items():
            if key.endswith(self.kind):
                if self.additive:
                    ctx[key] += bias
                    break  # only add to one thing, otherwise it's doubling up.
                else:
                    ctx[key] *= bias


@attr.s
class LogPoly(Foreground):
    """
    LogPoly model from Sims et.al. 2020 (Equation 18)
    T_{Fg} = 10^sum(d_i*log10(ν/ν_0)^i)_{i=0 to N}
    Parameters
    ----------
    poly_order : int
        The maximum polynomial order will be `poly_order `. There are `poly_order+1`
        total parameters.
    """

    poly_order = attr.ib(5, converter=int, kw_only=True)

    @cached_property
    def base_parameters(self):
        p = [
            Parameter("p0", 2, latex=r"p_0"),
        ]
        assert self.poly_order >= 1, "poly_order must be >= 1"

        # First create the parameters.
        for i in range(1, self.poly_order + 1):
            p.append(Parameter(f"p{i}", 0, latex=rf"p_{i}"))
        return tuple(p)

    @cached_property
    def logf(self):
        return np.log10(self.f)

    @cached_property
    def basis(self):
        return np.array([self.logf**i for i in range(self.poly_order + 1)])

    def model(self, **p):
        pp = [p[f"p{i}"] for i in range(self.poly_order + 1)]

        terms = [pp[i] * self.basis[i] for i in range(self.poly_order + 1)]
        return 10 ** np.sum(terms, axis=0)



"""Provides extra routines for fitting that are not in yabf."""
import numpy as np
from edges_cal.modelling import FixedLinearModel
from scipy import stats
from scipy.optimize import dual_annealing, minimize
from yabf import Component


class SemiLinearFit:
    def __init__(self, fg: FixedLinearModel, eor: Component, spectrum, sigma):
        """Perform a quick fit to data with a sum of linear and non-linear models.

        Useful for fitting foregrounds and EoR at the same time, where the EoR model is
        not linear, but the foreground model is.
        """

        self.fg = fg
        self.eor = eor
        self.spectrum = spectrum
        self.sigma = sigma

    def get_eor(self, p):
        return self.eor(params=p)["eor_spectrum"]

    def fg_fit(self, p):
        eor = self.get_eor(p)
        resid = self.spectrum - eor
        return self.fg.fit(
            ydata=resid,
            weights=1 / self.sigma**2 if hasattr(self.sigma, "__len__") else 1.0,
        )

    def fg_params(self, p):
        return self.fg_fit(p).model_parameters

    def get_resid(self, p):
        return self.fg_fit(p).residual

    def neg_lk(self, p):
        resid = self.get_resid(p)
        if hasattr(self.sigma, "ndim") and self.sigma.ndim == 2:
            norm_obj = stats.multivariate_normal(
                mean=np.zeros_like(resid), cov=self.sigma
            )
        else:
            norm_obj = stats.norm(loc=0, scale=self.sigma)

        return -np.sum(norm_obj.logpdf(resid))

    def __call__(self, dual_annealing_kw=None, **kwargs):
        if dual_annealing_kw is None:
            return minimize(
                self.neg_lk,
                x0=np.array([apar.fiducial for apar in self.eor.child_active_params]),
                bounds=[(apar.min, apar.max) for apar in self.eor.child_active_params],
                **kwargs
            )
        else:
            return dual_annealing(
                self.neg_lk,
                bounds=[(apar.min, apar.max) for apar in self.eor.child_active_params],
                x0=np.array([apar.fiducial for apar in self.eor.child_active_params]),
                local_search_options=kwargs,
                **dual_annealing_kw
            )



"""
This test module seeks to systematically test the NoiseWaveLikelihood likelihood.

It does this by first ensuring it gets reasonable results when the data is purely
simulated. It then goes on to test that if the input model is larger than the true
model, the extra parameters are estimated at zero. Then it goes on to compare the BIC
of each of these models, ensuring that the smaller model has a better BIC.
"""

import pytest

import numpy as np
from numdifftools import Hessian
from yabf import run_map

from edges_estimate.likelihoods import NoiseWaveLikelihood


@pytest.fixture(scope="module")
def fid_lk(calobs):
    return NoiseWaveLikelihood.from_calobs(calobs, seed=1234, as_sim="all")


@pytest.fixture(scope="module")
def larger_lk(calobs):
    return NoiseWaveLikelihood.from_calobs(calobs, as_sim="all", cterms=7, seed=1234)


@pytest.fixture(scope="module")
def true_tns(calobs):
    return calobs.C1_poly.coeffs[::-1] * calobs.t_load_ns


@pytest.fixture(scope="module")
def larger_tns(true_tns):
    return np.concatenate((true_tns, [0]))


@pytest.fixture(scope="module")
def fid_res(fid_lk):
    return run_map(fid_lk.partial_linear_model)


@pytest.fixture(scope="module")
def larger_res(larger_lk):
    return run_map(larger_lk.partial_linear_model)


def get_std(lk, res):
    h = Hessian(lambda x: lk.partial_linear_model(params=x)[0])
    return np.sqrt(np.diag(np.linalg.inv(-h(res.x))))


def test_pure_sim(fid_lk, fid_res, true_tns):
    assert fid_res.success
    std = get_std(fid_lk, fid_res)
    assert np.allclose(fid_res.x, true_tns, rtol=0, atol=3 * std)


def test_pure_sim_larger_model(larger_lk, larger_res, larger_tns):
    assert larger_res.success
    std = get_std(larger_lk, larger_res)
    assert np.allclose(larger_res.x, larger_tns, rtol=0, atol=3 * std)


def test_pure_sim_bic(fid_lk, fid_res, larger_lk, larger_res):
    fit, var, data = fid_lk.partial_linear_model.reduce_model(params=fid_res.x)
    nparams = fid_lk.partial_linear_model.linear_model.n_terms + len(
        fid_lk.partial_linear_model.child_active_params
    )
    fid_bic = nparams * len(data) - 2 * fid_lk.partial_linear_model(params=fid_res.x)[0]

    fit, var, data = larger_lk.partial_linear_model.reduce_model(params=larger_res.x)
    nparams = larger_lk.partial_linear_model.linear_model.n_terms + len(
        larger_lk.partial_linear_model.child_active_params
    )
    larger_bic = (
        nparams * len(data) - 2 * larger_lk.partial_linear_model(params=larger_res.x)[0]
    )

    assert fid_bic < larger_bic



import pytest

import attr
import matplotlib.pyplot as mpl
import numpy as np
from edges_cal.modelling import Polynomial
from yabf import Component, Parameter, run_map
from yabf.samplers.polychord import polychord

from edges_estimate.likelihoods import PartialLinearModel
from edges_estimate.plots import get_evidence


@attr.s(kw_only=True)
class Sin(Component):
    provides = ["power"]

    base_parameters = [
        Parameter("t", fiducial=0, min=-100, max=100, latex=r"t"),
    ]

    x: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))

    def calculate(self, ctx, **params):
        return params["t"] * np.sin(self.x * np.pi)


@attr.s(kw_only=True)
class Scale(Component):
    provides = ["scale"]

    base_parameters = [
        Parameter("scale", fiducial=1, min=0.1, max=100, latex=r"t"),
    ]

    x: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))

    def calculate(self, ctx, **params):
        # can't be scale*(anything) because then there's degeneracies with the other models.
        return 3 + params["scale"] * np.sin(self.x**2)


@pytest.fixture(scope="module")
def simple_plm():

    x = np.linspace(-1, 1, 100)

    linear_model = Polynomial(parameters=[1, 3]).at(x=x)
    sinx = Sin(x=x, params={"t": {"fiducial": 10, "min": -10, "max": 30}})

    y = linear_model() + sinx()["power"] + np.random.normal(scale=0.01, size=len(x))

    return PartialLinearModel(
        components=sinx,
        linear_model=linear_model,
        data={"data": y, "data_variance": 0.01},
        data_func=lambda ctx, data: data["data"] - ctx["power"],
    )


@pytest.fixture(scope="module")
def scaled_plm():

    x = np.linspace(-1, 1, 100)

    linear_model = Polynomial(parameters=[1, 3]).at(x=x)
    sinx = Sin(x=x, params={"t": {"fiducial": 10, "min": -10, "max": 30}})
    scale = Scale(x=x, params={"scale": {"fiducial": 1, "min": 0.1, "max": 30}})

    y = (
        linear_model() + sinx()["power"] + np.random.normal(scale=0.01, size=len(x))
    ) / scale()["scale"]

    def vfunc(ctx, data):
        v = ctx["scale"] ** 2 * data["data_variance"]
        return v

    return PartialLinearModel(
        components=(sinx, scale),
        linear_model=linear_model,
        data={
            "data": y,
            "data_variance": np.ones_like(y) * (0.01 / scale()["scale"]) ** 2,
        },
        data_func=lambda ctx, data: ctx["scale"] * data["data"] - ctx["power"],
        variance_func=vfunc,
    )


@pytest.fixture(scope="module")
def large_plm(simple_plm: PartialLinearModel):
    linear_model = Polynomial(parameters=[1, 3, 0]).at(x=simple_plm.linear_model.x)
    return attr.evolve(simple_plm, linear_model=linear_model)


@pytest.fixture(scope="module")
def scaled_large_plm(scaled_plm: PartialLinearModel):
    linear_model = Polynomial(parameters=[1, 3, 0]).at(x=scaled_plm.linear_model.x)
    return attr.evolve(scaled_plm, linear_model=linear_model)


def test_correct_result_simple(simple_plm):
    out = run_map(simple_plm)
    assert out.success
    np.testing.assert_allclose(out.x[0], 10, atol=0.02)

    fit, data, var = simple_plm.reduce_model(params=out.x)
    np.testing.assert_allclose(fit.residual, 0, atol=0.05)  # 5 sigma.


def test_correct_result_scaled(scaled_plm, plt):
    out = run_map(scaled_plm, x0=[10, 1])
    print(out)

    if plt == mpl:
        ctx = scaled_plm.get_ctx(params=[10, 1])
        fit, data, var = scaled_plm.reduce_model(ctx=ctx)

        ctx2 = scaled_plm.get_ctx(params=out.x)
        fit2, data2, var2 = scaled_plm.reduce_model(ctx=ctx2)
        fig, ax = plt.subplots(3, 2, sharex=True)
        x = scaled_plm.linear_model.x

        ax[0, 0].plot(scaled_plm.linear_model.x, scaled_plm.data["data"], label="data")
        ax[0, 0].plot(
            scaled_plm.linear_model.x,
            (fit.evaluate() + ctx["power"]) / ctx["scale"],
            label="model",
        )
        ax[0, 0].plot(
            scaled_plm.linear_model.x,
            (fit.evaluate()) / ctx["scale"],
            label="linear_model",
        )
        ax[0, 0].plot(
            scaled_plm.linear_model.x, (ctx["power"]) / ctx["scale"], label="sinx"
        )

        ax[1, 0].plot(scaled_plm.linear_model.x, data)
        ax[1, 0].plot(scaled_plm.linear_model.x, fit.evaluate())

        ax[2, 0].plot(scaled_plm.linear_model.x, fit.residual)
        ax[2, 0].fill_between(
            x, -np.sqrt(var) * np.ones_like(x), np.sqrt(var) * np.ones_like(x)
        )

        ax[0, 0].legend()

        ax[0, 1].plot(scaled_plm.linear_model.x, scaled_plm.data["data"], label="data")
        ax[0, 1].plot(
            scaled_plm.linear_model.x,
            (fit2.evaluate() + ctx2["power"]) / ctx2["scale"],
            label="model",
        )
        ax[0, 1].plot(
            scaled_plm.linear_model.x,
            (fit2.evaluate()) / ctx2["scale"],
            label="linear_model",
        )
        ax[0, 1].plot(
            scaled_plm.linear_model.x, (ctx2["power"]) / ctx2["scale"], label="sinx"
        )

        ax[1, 1].plot(scaled_plm.linear_model.x, data2)
        ax[1, 1].plot(scaled_plm.linear_model.x, fit2.evaluate())

        ax[2, 1].plot(scaled_plm.linear_model.x, fit2.residual)
        ax[2, 1].fill_between(
            x, -np.sqrt(var2) * np.ones_like(x), np.sqrt(var2) * np.ones_like(x)
        )

    assert out.success
    np.testing.assert_allclose(out.x[0], 10, atol=0.1)
    np.testing.assert_allclose(out.x[1], 1, atol=0.02)

    fit, data, v = scaled_plm.reduce_model(params=out.x)
    print(5 * np.sqrt(v))
    assert np.allclose(fit.residual, 0, atol=5 * np.sqrt(v))  # 5 sigma.


def test_correct_result_larger_model(large_plm):
    out = run_map(large_plm)

    np.testing.assert_allclose(out.x[0], 10, atol=0.05)
    fit, data, var = large_plm.reduce_model(params=out.x)
    np.testing.assert_allclose(fit.residual, 0, atol=0.05)  # 5 sigma.
    np.testing.assert_allclose(fit.model_parameters[-1], 0, rtol=0, atol=0.1)


def test_correct_result_larger_model_scaled(scaled_large_plm):
    out = run_map(scaled_large_plm)

    np.testing.assert_allclose(out.x[0], 10, atol=0.05)
    fit, data, var = scaled_large_plm.reduce_model(params=out.x)
    np.testing.assert_allclose(fit.residual, 0, atol=0.05)  # 5 sigma.
    np.testing.assert_allclose(fit.model_parameters[-1], 0, rtol=0, atol=0.1)


@pytest.mark.slow
def test_polychord_evidence(simple_plm, large_plm, tmp_path_factory):
    direc = tmp_path_factory

    poly = polychord(
        save_full_config=False,
        likelihood=simple_plm,
        output_dir=str(direc),
        output_prefix=f"{direc}/correct_model",
        sampler_kwargs={"nlive": 256, "precision_criterion": 0.1},
    )
    correct_samples = poly.sample()
    correct_lnz = get_evidence(correct_samples.root)

    poly = polychord(
        save_full_config=False,
        likelihood=large_plm,
        output_dir=str(direc),
        output_prefix=f"{direc}/larger_model",
        sampler_kwargs={"nlive": 256, "precision_criterion": 0.1},
    )
    large_samples = poly.sample()
    large_lnz = get_evidence(large_samples.root)
    assert large_lnz <= correct_lnz


@pytest.mark.slow
def test_polychord_evidence_scaled(scaled_plm, scaled_large_plm, tmp_path):
    direc = tmp_path / "cache"

    poly = polychord(
        save_full_config=False,
        likelihood=scaled_plm,
        output_dir=str(direc),
        output_prefix=f"{direc}/correct_scaled_model",
        sampler_kwargs={"nlive": 256, "precision_criterion": 0.1},
    )
    correct_samples = poly.sample()
    correct_lnz = get_evidence(correct_samples.root)

    poly = polychord(
        save_full_config=False,
        likelihood=scaled_large_plm,
        output_dir=str(direc),
        output_prefix=f"{direc}/larger_scaled_model",
        sampler_kwargs={"nlive": 256, "precision_criterion": 0.1},
    )
    large_samples = poly.sample()
    large_lnz = get_evidence(large_samples.root)
    assert large_lnz <= correct_lnz



import attr
import numpy as np
from yabf import Component, Parameter


def phenom_model(freqs, A, tau, w, nu0):
    """really bad inverse gaussian thing."""
    B = (
        4
        * (freqs - nu0) ** 2
        / w**2
        * np.log(-1 / tau * np.log((1 + np.exp(-tau)) / 2))
    )
    return -A * (1 - np.exp(-tau * np.exp(B))) / (1 - np.exp(-tau))


@attr.s
class AbsorptionProfile(Component):
    provides = ["eor_spectrum"]

    base_parameters = [
        Parameter("A", 0.5, min=0, latex=r"a_{21}"),
        Parameter("tau", 7, min=0, latex=r"\tau"),
        Parameter("w", 17.0, min=0),
        Parameter("nu0", 75, min=0, latex=r"\nu_0"),
    ]

    freqs: np.ndarray = attr.ib(kw_only=True, eq=attr.cmp_using(eq=np.array_equal))

    def calculate(self, ctx, **params):
        return phenom_model(self.freqs, **params)

    def spectrum(self, ctx, **params):
        return ctx["eor_spectrum"]



"""Top-level package for edges-estimate."""

from .calibration import *
from .eor_models import *
from .foregrounds import *
from .likelihoods import *

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

__author__ = """Steven Murray"""
__email__ = "steven.g.murray@asu.edu"
__version__ = version("edges_estimate")



import hickle
from astropy import units as u
from edges_cal import CalibrationObservation
from edges_io import io

calio = io.CalibrationObservation(
    "/data5/edges/data/CalibrationObservations/Receiver01/Receiver01_25C_2015_09_02_040_to_200MHz/",
    run_num={"receiver_reading": (1, 2, 4, 6, 8)},
    repeat_num=1,
)

calobs = CalibrationObservation.from_io(
    calio,
    f_low=50.0 * u.MHz,
    f_high=100.0 * u.MHz,
    cterms=6,
    wterms=5,
    spectrum_kwargs={
        "default": {"t_load": 300, "t_load_ns": 350, "ignore_times_percent": 7},
        "hot_load": {"ignore_times_percent": 10},
    },
    receiver_kwargs={"n_terms": 11, "model_type": "polynomial"},
)

hickle.dump(calobs, "data/test_calfile.h5")



from __future__ import annotations

import attr
import logging
import numpy as np
from astropy import units as u
from cached_property import cached_property
from edges_cal import CalibrationObservation
from edges_cal import receiver_calibration_func as rcf
from edges_cal import types as tp
from edges_cal.modelling import (
    CompositeModel,
    LinLog,
    Model,
    NoiseWaves,
    Polynomial,
    UnitTransform,
)
from edges_cal.simulate import simulate_q_from_calobs
from matplotlib import pyplot as plt
from pathlib import Path
from scipy import stats
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from typing import Callable, Sequence
from yabf import Component, Likelihood, Parameter, ParameterVector, ParamVec
from yabf.chi2 import Chi2, MultiComponentChi2

from .eor_models import AbsorptionProfile

logger = logging.getLogger(__name__)

try:
    profile  # noqa: F821
except NameError:

    def profile(fnc):
        return fnc


def _positive(x):
    assert x > 0


@attr.s(frozen=True)
class MultiComponentChi2SigmaLin(MultiComponentChi2):
    base_parameters = [
        Parameter("sigma_a", 0.013, min=0, latex=r"\sigma_a"),
        Parameter("sigma_b", 0.0, min=0, latex=r"\sigma_b"),
    ]

    nuc = attr.ib(75.0, converter=float, validator=_positive, kw_only=True)

    def get_sigma(self, model, **params):
        return (self.freqs / self.nuc) * params["sigma_b"] + params["sigma_a"]


@attr.s(frozen=True)
class MultiComponentChi2SigmaT(MultiComponentChi2):
    base_parameters = [
        Parameter("sigma_a", 0.013, min=0, latex=r"\sigma_a"),
        Parameter("sigma_b", 0.0, min=-1, max=1, latex=r"\sigma_b"),
    ]

    T0 = attr.ib(1750, kw_only=True, converter=float)

    def get_sigma(self, model, **params):
        return (model / self.T0) ** params["sigma_b"] * params["sigma_a"]


@attr.s(frozen=True)
class RadiometricAndWhiteNoise(MultiComponentChi2):
    """
    Likelihood with noise model based on Sims et al 2019 (1910.03165)

    This will only work if a single spectrum is used in the likelihood.

    Two tunable parameters exist: alpha_rn, the amplitude offset of the radiometric
    noise, and sigma_wn, the additive white-noise component.
    """

    base_parameters = [
        Parameter("alpha_rn", 1, min=0, max=100, latex=r"\alpha_{\rm rn}"),
        Parameter("sigma_wn", 0.0, min=0, latex=r"\sigma_{\rm wn}"),
    ]

    integration_time = attr.ib(converter=float, kw_only=True)  # in seconds!
    weights = attr.ib(1, kw_only=True)

    @weights.validator
    def _wght_validator(self, att, val):
        if type(val) == int and val == 1:
            return
        elif isinstance(val, np.ndarray) and val.shape == self.freqs.shape:
            return
        else:
            raise ValueError(
                f"weights must be an array with the same length as freqs."
                f"Got weight.shape == {val.shape} and freqs.shape == {self.freqs.shape}"
            )

    @cached_property
    def freqs(self):
        for cmp in self.components:
            if hasattr(cmp, "freqs"):
                return cmp.freqs

    @cached_property
    def channel_width(self):
        assert np.allclose(
            np.diff(self.freqs, 2), 0
        ), "the frequencies given are not regular!"
        return (self.freqs[1] - self.freqs[0]) * 1e6  # convert to Hz

    @cached_property
    def radiometer_norm(self):
        return self.channel_width * self.integration_time

    def get_sigma(self, model, **params):
        return np.sqrt(
            (1 / self.weights)
            * (
                params["alpha_rn"] * model**2 / self.radiometer_norm
                + params["sigma_wn"] ** 2
            )
        )


@attr.s(frozen=True, kw_only=True)
class PartialLinearModel(Chi2, Likelihood):
    r"""
    A likelihood where some of the parameters are linear and pre-marginalized.

    Parameters
    ----------
    linear_model
        A linear model containing all the terms that are linear.
    variance_func
        A callable function that takes two arguments: ``ctx`` and ``data``, and returns
        an array of model variance. If not provided, the input data must have a key
        called `"data_variance"` that provides static variance (i.e. the :math`\Sigma` in the
        derivation in the Notes).
    data_func
        A function that has the same signature as ``variance_func``, but returns data
        (i.e. the :math:`d` in the derivation). This might be dependent on non-linear
        parameters (not not the linear ones!). If not provided, the input data must have
        a key called ``"data"``.
    basis_func
        It is not recommended to provide this, but if provided it should be a function
        that takes the linear basis, context and data, and returns a new linear model,
        effectively altering the linear basis functions based on the nonlinear parameters.

    Notes
    -----
    The general idea is laid out in Monsalve et al. 2018
    (or https://arxiv.org/pdf/0809.0791.pdf). In this class, the variables are typically
    named the same as in Monsalve et al. 2018 (eg. Q, C, V, Sigma).
    However, note that M2018 misses some fairly significant simplifications.

    Eq. 15 of M18 is

    .. math:: d_\star^T \Sigma^{-1} d_\start - d_\star^T \Sigma^{-1} A (A^T \Sigma^{-1} A)^{-1} A^T \Sigma^{-1} d_\star

    where :math:`d_\star`  is the residual of the linear model: :math:`d_star = d - A\hat{\theta}`.
    (note that we're omitting the nonlinear 21cm model from that paper here because
    it's just absorbed into :math:`d`.) Note that part of the second term is just the
    "hat" matrix from weighted least-squares, i.e.

    .. math:: H = A (A^T \Sigma^{-1} A)^{-1} A^T \Sigma^{-1}

    which when applied to a data vector, returns the maximum-likelihood model of the data.

    Thus we can re-write

    .. math:: d_\star^T \Sigma^{-1} d_\start - d_\star^T \Sigma^{-1} H (d - A \hat{\theta}).

    But :math:`A\hat{\theta}` is equivalent to `H d` (i.e. both produce the maximum
    likelihood model for the data), so we have

    .. math:: d_\star^T \Sigma^{-1} d_\start - d_\star^T \Sigma^{-1} H (d - Hd).

    But the `H` matrix is idempotent, so :math:`Hd - HHd = Hd - Hd = 0`. So we are left
    with the first term only.
    """
    linear_model: Model = attr.ib()
    variance_func: Callable | None = attr.ib(default=None)
    data_func: Callable | None = attr.ib(default=None)
    basis_func: Callable | None = attr.ib(default=None)
    subtract_fiducial: bool = attr.ib(default=False)
    verbose: bool = attr.ib(False)

    @profile
    def _reduce(self, ctx, **params):
        if self.variance_func is None:
            var = self.data["data_variance"]
        else:
            var = self.variance_func(ctx, self.data)

        if self.data_func is None:
            data = self.data["data"]
        else:
            data = self.data_func(ctx, self.data)

        if self.basis_func is None:
            linear_model = self.linear_model
        else:
            linear_model = self.basis_func(self.linear_model, ctx, self.data)

        wght = 1.0 if np.all(var == 0) else 1 / var

        linear_fit = linear_model.fit(ydata=data, weights=wght)
        return linear_fit, data, var

    @cached_property
    def Q(self):
        if self.basis_func is not None or self.variance_func is not None:
            raise AttributeError("Q is not static in this instance!")
        return (self.linear_model.basis / self.data["data_variance"]).dot(
            self.linear_model.basis.T
        )

    @cached_property
    def logdetCinv(self) -> float | None:
        if np.all(self.data["data_variance"] == 0):
            return 0.0

        try:
            Cinv = self.Q
            return np.log(np.linalg.det(Cinv))
        except AttributeError:
            return None

    @cached_property
    def sigma_plus_v_inverse(self):
        if self.basis_func is not None or self.variance_func is not None:
            raise AttributeError("V is not static in this instance!")
        A = self.linear_model.basis
        var = self.data["data_variance"]
        Sig = np.diag(var)
        SigInv = np.diag(1 / var)
        C = np.linalg.inv(self.Q)
        SigFG = A.T.dot(C.dot(A))
        V = np.linalg.inv(np.linalg.inv(SigFG) - SigInv)
        return np.linalg.inv(Sig + V)

    @cached_property
    def fiducial_lnl(self):
        return attr.evolve(self, subtract_fiducial=False, verbose=False)()[0]

    @profile
    def logdet_cinv(self, model, ctx, **params) -> float:
        """A derived quantity, the log-determinant of the inverse of C."""
        if self.logdetCinv is None:
            fit = model[0]
            h = fit.hessian
            return np.log(np.linalg.det(h))
        else:
            return self.logdetCinv

    def logdet_sig(self, model, ctx, **params):
        """A derived quantity, the log-determinant of the covariance matrix."""
        var = model[-1]

        if not hasattr(var, "__len__"):
            var = var * np.ones(len(model[1]))
        elif np.all(var == 0):
            var = np.ones_like(var)

        var = var[~np.isinf(var)]

        return np.sum(np.log(var)) if self.variance_func is not None else 0

    def rms(self, model, ctx, **params):
        fit, data, var = model

        if not hasattr(var, "__len__"):
            var = var * np.ones(len(data))
        elif np.all(var == 0):
            var = np.ones_like(var)

        mask = ~np.isinf(var)
        data = data[mask]
        resid = fit.residual[mask]
        var = var[mask]

        return np.nansum(resid**2 / var)

    @profile
    def lnl(self, model, **params):
        # Ensure we don't use flagged channels
        logdetSig = self.logdet_sig(model, None, **params)
        logdetCinv = self.logdet_cinv(model, None, **params)
        rms = self.rms(model, None, **params)

        lnl = -0.5 * (logdetSig + logdetCinv + rms)

        if np.isnan(lnl):
            lnl = -np.inf

        if self.subtract_fiducial:
            lnl -= self.fiducial_lnl

        if self.verbose:
            print(params, lnl)

        if np.isnan(lnl) or np.isinf(lnl):
            logger.warn(f"Got bad log-likelihood: {lnl} for params: {params}")

        return lnl

    def get_unmarginalized_lnl(self, linear_params, nonlinear_params):
        ctx = self.get_ctx(params=nonlinear_params)

        if self.variance_func is None:
            var = self.data["data_variance"]
        else:
            var = self.variance_func(ctx, self.data)

        if self.data_func is None:
            data = self.data["data"]
        else:
            data = self.data_func(ctx, self.data)

        if self.basis_func is None:
            linear_model = self.linear_model
        else:
            linear_model = self.basis_func(self.linear_model, ctx, self.data)

        linear = linear_model(parameters=linear_params)

        resid = data - linear
        nm = stats.norm(loc=0, scale=np.sqrt(var))
        return np.sum(nm.logpdf(resid))

    def _extracted_from_lnl_45(self, basis, var):
        A = basis
        C = np.linalg.inv(A.T.dot(A / var))
        result = np.linalg.inv(A.T.dot(C.dot(A)))
        result[np.diag_indices_from(result)] -= 1 / var
        result = np.linalg.inv(result)
        result[np.diag_indices_from(result)] += var
        result = np.linalg.inv(result)

        return result


@attr.s(frozen=True, kw_only=True)
class TNS(Component):
    x: np.ndarray = attr.ib()
    c_terms: int = attr.ib(default=5)
    field_freq: np.ndarray | None = attr.ib(None)
    freq_range: tuple[float, float] = attr.ib()
    fiducial: np.typing.ArrayLike = attr.ib()

    @freq_range.default
    def _fr_default(self):
        return (self.x.min(), self.x.max())

    @freq_range.validator
    def _fr_validator(self, att, val):
        if len(val) != 2:
            raise ValueError("freq_range must be a 2-tuple of floats!")
        if val[1] <= val[0]:
            raise ValueError(f"freq_range must in the form (min, max), got: {val}")

    @fiducial.default
    def _fid_default(self):
        return [1500] + [0] * (self.c_terms - 1)

    @fiducial.validator
    def _fid_validator(self, att, val):
        if len(val) != self.c_terms:
            raise ValueError(
                f"fiducial must be a sequence of length c_terms ({self.c_terms}), got {len(val)}."
            )

    @cached_property
    def provides(self):
        if self.field_freq is None:
            return ["tns"]
        else:
            return ["tns", "tns_field"]

    @cached_property
    def base_parameters(self):
        return ParameterVector(
            "t_lns",
            fiducial=self.fiducial,
            length=self.c_terms,
            latex=r"T^{\rm L+NS}_{%s}",
        ).get_params()

    @cached_property
    def model(self):
        return Polynomial(
            n_terms=self.c_terms,
            transform=UnitTransform(range=self.freq_range),
            parameters=[p.fiducial for p in self.active_params],
        ).at(x=self.x)

    @cached_property
    def field_model(self):
        return Polynomial(
            n_terms=self.c_terms,
            transform=UnitTransform(range=self.freq_range),
            parameters=[p.fiducial for p in self.active_params],
        ).at(x=self.field_freq)

    def calculate(self, ctx, **params):

        tns = self.model(parameters=list(params.values()))

        if self.field_freq is not None:
            tns_field = self.field_model(parameters=list(params.values()))
            return tns, tns_field
        else:
            return tns


@attr.s(frozen=True, kw_only=True)
class S11Systematic(Component):
    freq: np.ndarray = attr.ib()
    measured: np.ndarray = attr.ib()
    _scale_model: Model = attr.ib()
    _delay_model: Model = attr.ib()

    @cached_property
    def scale_model(self):
        return self._scale_model.at(x=self.freq)

    @cached_property
    def delay_model(self):
        return self._delay_model.at(x=self.freq)

    @cached_property
    def base_parameters(self):
        return tuple(
            Parameter(f"logscale_{i}", fiducial=0, latex=rf"A^\Gamma_{i}")
            for i in range(self.scale_model.n_terms)
        ) + tuple(
            Parameter(f"delay_{i}", fiducial=0, latex=rf"tau^\Gamma_{i}")
            for i in range(self.delay_model.n_terms)
        )

    @cached_property
    def provides(self):
        return {f"{self.name}_gamma"}

    def calculate(self, ctx, **params):
        s = [params[f"logscale_{i}"] for i in range(self.scale_model.n_terms)]
        d = [params[f"delay_{i}"] for i in range(self.delay_model.n_terms)]

        return (self.measured * 10 ** self.scale_model(parameters=s)) * np.exp(
            -1j * self.delay_model(parameters=d) / 1000 * self.freq
        )


@attr.s(frozen=True, kw_only=True)
class NoiseWaveCoefficients(Component):
    # Requires two sub-components: a source s11 and an LNA s11
    source_names = attr.ib()
    as_dict = attr.ib(False)

    @cached_property
    def provides(self):
        return {"K"}

    @profile
    def calculate(self, ctx, **params):
        if not self.as_dict:
            return np.hstack(
                tuple(
                    rcf.get_K(
                        gamma_rec=ctx["rcv_gamma" if src != "ant" else "rcvant_gamma"],
                        gamma_ant=ctx[f"{src}_gamma"],
                    )
                    for src in self.source_names
                )
            )
        else:
            return {
                src: rcf.get_K(
                    gamma_rec=ctx["rcv_gamma" if src != "ant" else "rcvant_gamma"],
                    gamma_ant=ctx[f"{src}_gamma"],
                )
                for src in self.source_names
            }


@attr.s(frozen=True, kw_only=True)
class NoiseWaveLikelihood:
    nw_model: NoiseWaves = attr.ib()
    data: dict = attr.ib()
    sig_by_tns: bool = attr.ib(default=True)
    t_ns_params: ParamVec = attr.ib()
    t_ns_freq_range: tuple[float, float] = attr.ib()
    derived: tuple[str] = attr.ib(factory=tuple)
    s11_systematics: Sequence[S11Systematic] = attr.ib(())

    @t_ns_freq_range.default
    def _tns_fr_default(self):
        return (self.nw_model.freq.min(), self.nw_model.freq.max())

    @t_ns_params.default
    def _tns_default(self) -> ParamVec:
        return ParamVec(
            "t_lns",
            length=self.nw_model.c_terms,
            fiducial=[1000] + [0] * (self.nw_model.c_terms - 1),
        )

    @cached_property
    def t_ns_model(self):
        if hasattr(self.nw_model.freq, "unit"):
            f = self.nw_model.freq.to_value("MHz")
        else:
            f = self.nw_model.freq
        return TNS(
            x=f,
            c_terms=self.nw_model.c_terms,
            params=self.t_ns_params.get_params(),
            fiducial=list(self.t_ns_params.fiducial),
            freq_range=self.t_ns_freq_range,
        )

    @cached_property
    def partial_linear_model(self):
        return PartialLinearModel(
            linear_model=self.nw_model.linear_model,
            data=self.data,
            components=(self.t_ns_model,) + self.s11_systematics,
            data_func=self.transform_data,
            variance_func=self.transform_variance if self.sig_by_tns else None,
            derived=tuple(
                d if hasattr(PartialLinearModel, d) else getattr(self, d)
                for d in self.derived
            ),
            basis_func=self.apply_s11_systematics if self.s11_systematics else None,
        )

    @classmethod
    def transform_data(cls, ctx: dict, data: dict):
        n = len(data["q"]) // len(ctx["tns"])
        tns = np.concatenate((ctx["tns"],) * n)
        k0 = ctx["K"][0] if "K" in ctx else data["k0"]
        return data["q"] * tns - k0 * data["T"]

    @classmethod
    def transform_variance(cls, ctx: dict, data: dict):
        n = len(data["q"]) // len(ctx["tns"])
        tns = np.concatenate((ctx["tns"],) * n)
        return data["data_variance"] * tns**2

    @classmethod
    @profile
    def apply_s11_systematics(cls, linear_model, ctx, data):
        K = ctx["K"]

        extra_basis = {"tunc": K[1], "tcos": K[2], "tsin": K[3]}
        m = linear_model.model
        repeater = (m.tunc.n_terms, m.tcos.n_terms, m.tsin.n_terms)
        if "tload" in m.extra_basis:
            extra_basis["tload"] = m.extra_basis["tload"]
            repeater = repeater + (m.tload.n_terms,)

        # We do the following (updating the basis functions manually) for speed
        # If we don't pass "init_basis" explicitly, the whole basis function has to
        # be recomputed, whereas if we just multiply the new/old K, we get a decent
        # speedup.
        K = np.vstack((K[1:], np.atleast_2d(m.extra_basis["tload"])))
        new_extra_basis = np.repeat(K, repeater, axis=0)

        return attr.evolve(
            linear_model,
            init_basis=data["raw_basis"] * new_extra_basis,
            model=attr.evolve(linear_model.model, extra_basis=extra_basis),
        )

    @classmethod
    def from_calobs(
        cls,
        calobs,
        cterms=None,
        wterms=None,
        sig_by_sigq=True,
        sources=None,
        as_sim=None,
        reweight_sources: bool = False,
        reweight_frequencies: bool = False,
        s11_systematic_params=None,
        cable_noise_factor=1,
        source_factors=None,
        include_antsim: bool = False,
        seed: int = 1234,
        add_noise: bool = True,
        **kwargs,
    ):
        if sources is None:
            sources = tuple(calobs.loads.keys())

        s11_systematic_params = s11_systematic_params or {}

        as_sim = as_sim or []
        if as_sim == "all":
            as_sim = list(calobs.loads.keys())

        loads = {src: load for src, load in calobs.loads.items() if src in sources}
        if include_antsim:
            for name in calobs.metadata["io"].s11.simulators:
                loads[name] = calobs.new_load(name, io_obj=calobs.metadata["io"])

        sources = tuple(loads.keys())

        # Source Factors define how much "extra" weight to place on each source.
        if source_factors is None:
            source_factors = (1,) * len(loads)

        nw_model = NoiseWaves.from_calobs(
            calobs, loads=loads, cterms=cterms, wterms=wterms
        )

        raw_bases = nw_model.get_linear_model(with_k=False).basis

        freq = calobs.freq.freq.to_value("MHz")
        if s11_systematic_params:
            lna_delay_model = s11_systematic_params.get("rcv", {}).pop(
                "delay_model", Polynomial(n_terms=1)
            )
            lna_scale_model = s11_systematic_params.get("rcv", {}).pop(
                "scale_model", Polynomial(n_terms=1)
            )

            lna = S11Systematic(
                freq=freq,
                measured=calobs.receiver.s11_model(freq),
                params=s11_systematic_params.get("rcv", {}),
                delay_model=lna_delay_model,
                scale_model=lna_scale_model,
                name="rcv",
            )

            s11_systematics = tuple(
                S11Systematic(
                    freq=freq,
                    measured=load.reflections.s11_model(freq),
                    params=s11_systematic_params.get(src, {}),
                    delay_model=s11_systematic_params.get(src, {}).pop(
                        "delay_model", Polynomial(n_terms=1)
                    ),
                    scale_model=s11_systematic_params.get(src, {}).pop(
                        "scale_model", Polynomial(n_terms=1)
                    ),
                    name=src,
                )
                for src, load in loads.items()
            )

            noise_wave_coeffs = (
                NoiseWaveCoefficients(
                    source_names=sources, components=s11_systematics + (lna,)
                ),
            )
        else:
            noise_wave_coeffs = ()

        ks = {
            name: load.reflections.get_k_matrix(calobs.receiver, freq=nw_model.freq)
            for name, load in loads.items()
        }
        k0 = np.concatenate(tuple(ks[src][0] for src in loads))

        if as_sim:
            np.random.seed(seed)

        data = {
            "q": np.concatenate(
                tuple(
                    simulate_q_from_calobs(calobs, name)
                    + (
                        np.random.normal(
                            scale=np.sqrt(
                                load.spectrum.variance_Q / load.spectrum.n_integrations
                            )
                        )
                        if add_noise
                        else 0
                    )
                    if name in as_sim
                    else load.spectrum.averaged_Q
                    for name, load in loads.items()
                )
            ),
            "T": np.concatenate(
                tuple(load.temp_ave * np.ones(calobs.freq.n) for load in loads.values())
            ),
            "k0": k0,
            "gamma_src": {
                name: source.s11_model(freq) for name, source in loads.items()
            },
            "gamma_rcv": calobs.receiver.s11_model(freq),
            "raw_basis": raw_bases,
        }

        overall_mean = np.mean(
            [np.mean(load.spectrum.variance_Q) for load in loads.values()]
        )

        if sig_by_sigq:
            if reweight_sources:
                data["data_variance"] = np.concatenate(
                    tuple(
                        overall_mean
                        * load.spectrum.variance_Q
                        / np.mean(load.spectrum.variance_Q)
                        for load in loads.values()
                    )
                )
            elif reweight_frequencies:
                data["data_variance"] = np.concatenate(
                    tuple(
                        np.mean(load.spectrum.variance_Q)
                        * np.ones_like(load.spectrum.variance_Q)
                        for load in loads.values()
                    )
                )
            else:
                data["data_variance"] = np.concatenate(
                    tuple(
                        load.spectrum.variance_Q / load.spectrum.n_integrations
                        for load in loads.values()
                    )
                )
        else:
            data["data_variance"] = np.concatenate(
                tuple(
                    overall_mean * np.ones_like(load.spectrum.variance_Q)
                    for load in loads.values()
                )
            )

        noise_factor = np.concatenate(
            tuple(
                np.ones_like(load.spectrum.averaged_Q)
                * (cable_noise_factor if name in ["open", "short"] else 1)
                * source_factors[i]
                for i, (name, load) in enumerate(loads.items())
            )
        )

        data["data_variance"] *= noise_factor / np.mean(noise_factor)

        return cls(
            nw_model=nw_model, data=data, s11_systematics=noise_wave_coeffs, **kwargs
        )

    def get_cal_curves(
        self, params: Sequence | None = None, freq=None, sample=True
    ) -> dict[str, np.ndarray]:
        fit = self.partial_linear_model.reduce_model(params=params)[0]

        if freq is None:
            freq = self.nw_model.freq

        model = fit.fit.model
        if sample:
            linear_params = fit.get_sample()[0]
        else:
            linear_params = fit.model_parameters

        out = {
            name: model.get_model(name, x=freq, parameters=linear_params)
            for name in model.models
        }

        pp = []
        for i, p in enumerate(self.partial_linear_model.child_active_params):
            if p.name.startswith("t_lns"):
                pp.append(params[i])

        out["tns"] = self.t_ns_model.model.model(x=freq, parameters=pp)
        out["params"] = linear_params
        return out

    def get_linear_coefficients(
        self,
        freq,
        labcal,
        params=None,
        ctx=None,
        fit=None,
        linear_params=None,
        load=None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Get the linear coefficients that convert uncalibrated to calibrated temp.

        The equation is T_cal = a*T_uncal + b
        """
        pp = []
        for i, p in enumerate(self.partial_linear_model.child_active_params):
            if p.name.startswith("t_lns"):
                pp.append(params[i])
        tns = self.t_ns_model.model.model(x=freq, parameters=pp)

        model = self.nw_model.linear_model.model
        if linear_params is not None:
            params = linear_params
        else:
            ctx = self.partial_linear_model.get_ctx(params=params)
            if fit is None:
                fit = self.partial_linear_model.reduce_model(ctx=ctx)[0]
            params = fit.model_parameters

        if load is None:
            gamma_ant = labcal.antenna_s11_model(freq)
        elif load.load_name + "_gamma" in ctx:
            gamma_ant = ctx[load.load_name + "_gamma"]
        else:
            gamma_ant = load.reflections.s11_model(freq)

        if "rcv_gamma" in ctx:
            rcv = ctx["rcv_gamma"]

            if len(freq) != len(rcv):
                re = spline(self.nw_model.freq, np.real(rcv))(freq)
                im = spline(self.nw_model.freq, np.imag(rcv))(freq)
                rcv = re + 1j * im
        else:
            rcv = labcal.calobs.receiver_s11(freq * u.MHz)

        return rcf.get_linear_coefficients(
            gamma_ant=gamma_ant,
            gamma_rec=rcv,
            sca=tns / labcal.calobs.t_load_ns,
            off=labcal.calobs.t_load
            - model.get_model("tload", x=freq, parameters=params),
            t_unc=model.get_model("tunc", x=freq, parameters=params),
            t_cos=model.get_model("tcos", x=freq, parameters=params),
            t_sin=model.get_model("tsin", x=freq, parameters=params),
            t_load=labcal.calobs.t_load,
        )

    # Start derived parameters
    @staticmethod
    def tunchat(model, ctx, **params):
        return model[0].fit.model.tunc.parameters

    @staticmethod
    def tcoshat(model, ctx, **params):
        return model[0].fit.model.tcos.parameters

    @staticmethod
    def tsinhat(model, ctx, **params):
        return model[0].fit.model.tsin.parameters

    @staticmethod
    def tloadhat(model, ctx, **params):
        return model[0].fit.model.tload.parameters

    @staticmethod
    def linear(model, ctx, **params):
        fit = model[0]
        return fit.get_sample()

    def rms_parts(self, model, ctx, **params):
        fit, data, var = model

        rms_parts = []
        n = 0
        for i, (name, d) in enumerate(self.data.items()):
            v = var[n : n + len(d)]
            r = fit.residual[n : n + len(d)]

            mask = (~np.isinf(v)) & (~np.isnan(v))

            rms_parts.append(np.nansum(r[mask] ** 2 / v[mask]))
            n += len(d)

        return rms_parts

    def lnl_at_calobs(self, calobs: CalibrationObservation):
        """Compute the (unmarginalized) likelihood at the same parameters as a calobs."""
        p = np.concatenate(
            (
                calobs.Tunc_poly.coeffs[::-1],
                calobs.Tcos_poly.coeffs[::-1],
                calobs.Tsin_poly.coeffs[::-1],
                calobs.t_load - calobs.C2_poly.coeffs[::-1],
            )
        )

        nl = calobs.C1_poly.coeffs[::-1] * calobs.t_load_ns

        return self.partial_linear_model.get_unmarginalized_lnl(p, nl)


@attr.s(frozen=True, kw_only=True)
class NoiseWavesPlusFG:
    freq: tp.FreqType = attr.ib()
    _gamma_src: dict[str, np.ndarray] = attr.ib()
    gamma_ant: np.ndarray = attr.ib()
    gamma_rec: Callable = attr.ib()
    field_freq: tp.FreqType = attr.ib()
    c_terms: int = attr.ib(default=5)
    w_terms: int = attr.ib(default=6)
    fg_model: Model = attr.ib(default=LinLog(n_terms=5))
    parameters: Sequence[float] | None = attr.ib(default=None)
    loss: float | np.ndarray = attr.ib(default=1.0)
    bm_corr: float | np.ndarray = attr.ib(default=1.0)

    @field_freq.default
    def _ff_default(self) -> np.ndarray:
        return self.freq

    @cached_property
    def gamma_src(self):
        return {**self._gamma_src, **{"ant": self.gamma_ant}}

    def _freq(self, src: str):
        if src == "ant":
            return self.field_freq
        else:
            return self.freq

    @cached_property
    def src_names(self) -> list[str]:
        """List of names of inputs sources (eg. ambient, hot_load, open, short)."""
        return list(self.gamma_src.keys())

    @cached_property
    def linear_model(self):
        return self.get_linear_model()

    def get_linear_model(self, with_k: bool = True) -> CompositeModel:
        """The actual composite linear model object associated with the noise waves."""
        # K should be a an array of shape (Nsrc Nnu x Nnoisewaveterms)
        if with_k:
            K = np.hstack(
                tuple(
                    rcf.get_K(
                        gamma_rec=self.gamma_rec(self._freq(name)),
                        gamma_ant=self.gamma_src[name](self._freq(name)),
                    )
                    for name in self.src_names
                )
            )
            # K[0] multiples the fg, but not the other models.
            K[0][: len(self.freq) * (len(self.gamma_src) - 1)] = 0.0
            K[0][-len(self.field_freq) :] *= self.loss * self.bm_corr

        x = np.concatenate(
            (
                np.tile(self.freq.to_value("MHz"), len(self.src_names) - 1),
                self.field_freq.to_value("MHz"),
            )
        )

        transform = UnitTransform(
            range=[self.freq.to_value("MHz").min(), self.freq.to_value("MHz").max()]
        )

        models = {
            "tunc": Polynomial(
                n_terms=self.w_terms,
                parameters=self.parameters[: self.w_terms]
                if self.parameters is not None
                else None,
                transform=transform,
            ),
            "tcos": Polynomial(
                n_terms=self.w_terms,
                parameters=self.parameters[self.w_terms : 2 * self.w_terms]
                if self.parameters is not None
                else None,
                transform=transform,
            ),
            "tsin": Polynomial(
                n_terms=self.w_terms,
                parameters=self.parameters[2 * self.w_terms : 3 * self.w_terms]
                if self.parameters is not None
                else None,
                transform=transform,
            ),
            "tload": Polynomial(
                n_terms=self.c_terms,
                parameters=(
                    self.parameters[3 * self.w_terms : 3 * self.w_terms + self.c_terms]
                    if self.parameters is not None
                    else None
                ),
                transform=transform,
            ),
            "fg": self.fg_model,
        }

        if with_k:
            extra_basis = {
                "tunc": K[1],
                "tcos": K[2],
                "tsin": K[3],
                "tload": -1 * np.ones(len(x)),
                "fg": K[0],
            }
        else:
            extra_basis = {}

        return CompositeModel(models=models, extra_basis=extra_basis).at(x=x)

    def _get_idx(self, src: str):
        if src == "ant":
            return slice(-len(self.field_freq), None, None)
        else:
            return slice(
                self.src_names.index(src) * len(self.freq),
                (1 + self.src_names.index(src)) * len(self.freq),
                None,
            )

    def get_temperature_term(
        self,
        noise_wave: str,
        parameters: Sequence | None = None,
        src: str | None = None,
    ) -> np.ndarray:
        """Get the model for a particular temperature term."""
        out = self.linear_model.model.get_model(
            noise_wave,
            parameters=parameters,
            x=self.linear_model.x,
            with_extra=bool(src),
        )

        if src:
            return out[self._get_idx(src)]
        else:
            return out[: len(self.linear_model.x)]

    def get_full_model(
        self, src: str, parameters: Sequence | None = None
    ) -> np.ndarray:
        """Get the full model (all noise-waves) for a particular input source."""
        out = self.linear_model(parameters=parameters)
        return out[self._get_idx(src)]

    def get_fitted(
        self, data: np.ndarray, weights: np.ndarray | None = None
    ) -> NoiseWaves:
        """Get a new noise wave model with fitted parameters."""
        fit = self.linear_model.fit(ydata=data, weights=weights)
        return attr.evolve(self, parameters=fit.model_parameters)

    def with_params_from_calobs(self, calobs, cterms=None, wterms=None, fg_terms=None):
        cterms = cterms or calobs.cterms
        wterms = wterms or calobs.wterms

        def modify(thing, n):
            if len(thing) < wterms:
                return thing + [0] * (n - len(thing))
            elif len(thing) > wterms:
                return thing[:n]
            else:
                return thing

        tu = modify(calobs.Tunc_poly.coefficients[::-1].tolist(), wterms)
        tc = modify(calobs.Tcos_poly.coefficients[::-1].tolist(), wterms)
        ts = modify(calobs.Tsin_poly.coefficients[::-1].tolist(), wterms)

        c2 = (-calobs.C2_poly.coefficients[::-1]).tolist()
        c2[0] += calobs.t_load
        c2 = modify(c2, cterms)

        if fg_terms is None:
            fg_terms = [0] * self.fg_model.n_terms

        return attr.evolve(self, parameters=tu + tc + ts + c2 + list(fg_terms))

    @classmethod
    def from_labcal(
        cls,
        labcal,
        calobs,
        fg_model=LinLog(n_terms=5),
        loads: dict | None = None,
        cterms=None,
        wterms=None,
        **kwargs,
    ) -> NoiseWavesPlusFG:
        """Initialize a noise wave model from a calibration observation."""
        cterms = cterms or calobs.cterms
        wterms = wterms or calobs.wterms

        if loads is None:
            loads = calobs.loads

        gamma_src = {name: load.reflections.s11_model for name, load in loads.items()}

        out = cls(
            freq=calobs.freq.freq,
            gamma_src=gamma_src,
            gamma_rec=calobs.receiver.s11_model,
            gamma_ant=labcal.antenna_s11_model,
            c_terms=cterms,
            w_terms=wterms,
            fg_model=fg_model,
            **kwargs,
        )
        return out.with_params_from_calobs(calobs, cterms, wterms, fg_model.parameters)

    def __call__(self, **kwargs) -> np.ndarray:
        """Call the underlying linear model."""
        return self.linear_model(**kwargs)


@attr.s(frozen=True, kw_only=True)
class DataCalibrationLikelihood:
    nwfg_model = attr.ib()
    data: dict = attr.ib()
    eor_components = attr.ib()
    t_ns_params: ParamVec | None = attr.ib(None)
    verbose: bool = attr.ib(False)
    subtract_fiducial: bool = attr.ib(False)
    save_linear_params: bool = attr.ib(True)
    save_sampled_linear_params: bool = attr.ib(True)
    derived: list[str | Callable] = attr.ib(factory=list)
    s11_systematics: Sequence[S11Systematic] = attr.ib(())

    @eor_components.default
    def _eorcmp(self):
        return (
            AbsorptionProfile(
                freqs=self.nwfg_model.freq, params=("A", "w", "tau", "nu0")
            ),
        )

    @cached_property
    def src_names(self) -> tuple[str]:
        return tuple(self.data["q"].keys())

    @cached_property
    def t_ns_model(self):
        if self.t_ns_params is None:
            t_ns_params = ParamVec("t_lns", length=self.nwfg_model.c_terms)
        else:
            t_ns_params = self.t_ns_params
        return TNS(
            x=self.nwfg_model.freq.to_value("MHz"),
            field_freq=self.nwfg_model.field_freq,
            c_terms=self.nwfg_model.c_terms,
            params=t_ns_params.get_params(),
        )

    # Start derived parameters
    @staticmethod
    def tunchat(model, ctx, **params):
        return model[0].fit.model.tunc.parameters

    @staticmethod
    def tcoshat(model, ctx, **params):
        return model[0].fit.model.tcos.parameters

    @staticmethod
    def tsinhat(model, ctx, **params):
        return model[0].fit.model.tsin.parameters

    @staticmethod
    def tloadhat(model, ctx, **params):
        return model[0].fit.model.tload.parameters

    @staticmethod
    def tfghat(model, ctx, **params):
        return model[0].fit.model.fg.parameters

    @staticmethod
    def linear(model, ctx, **params):
        fit = model[0]
        return fit.get_sample()

    @cached_property
    def partial_linear_model(self):
        derived = []
        if self.save_linear_params:
            derived.extend(["tunchat", "tcoshat", "tsinhat", "tloadhat", "tfghat"])
        if self.save_sampled_linear_params:
            derived.append("linear")

        derived.extend(self.derived)

        return PartialLinearModel(
            linear_model=self.nwfg_model.linear_model,
            data=self.data,
            components=(self.t_ns_model,) + self.eor_components + self.s11_systematics,
            data_func=self.transform_data,
            variance_func=self.transform_variance,
            verbose=self.verbose,
            subtract_fiducial=self.subtract_fiducial,
            derived=tuple(
                d if hasattr(PartialLinearModel, d) else getattr(self, d)
                for d in derived
            ),
            basis_func=self.apply_s11_systematics if self.s11_systematics else None,
        )

    def transform_data(self, ctx: dict, data: dict):
        tns = ctx["tns"]
        Tant = ctx["eor_spectrum"]
        if "K" in ctx:
            k0 = {name: kk[0] for name, kk in ctx["K"].items()}
        else:
            k0 = data["k0"]

        out = []
        for src in self.src_names:
            if src == "ant":
                out.append(
                    ctx["tns_field"] * data["q"]["ant"]
                    - k0["ant"]
                    * (Tant * data["loss"] * data["bm_corr"] + data["loss_temp"])
                )
            else:
                Tsrc = data["T"][src]
                out.append(tns * data["q"][src] - k0[src] * Tsrc)

        return np.concatenate(out)

    def transform_variance(self, ctx: dict, data: dict):
        tns = ctx["tns"]
        field_tns = ctx["tns_field"]

        return np.concatenate(
            [
                data["data_variance"][src]
                * (field_tns**2 if src == "ant" else tns**2)
                for src in self.src_names
            ]
        )

    @classmethod
    @profile
    def apply_s11_systematics(cls, linear_model, ctx, data):
        K = ctx["K"]
        m = linear_model.model

        extra_basis = {
            "tunc": np.concatenate([kk[1] for kk in K.values()]),
            "tcos": np.concatenate([kk[2] for kk in K.values()]),
            "tsin": np.concatenate([kk[3] for kk in K.values()]),
            "tload": m.extra_basis["tload"],
            "fg": np.concatenate([kk[0] for kk in K.values()]),
        }
        repeater = (
            m.tunc.n_terms,
            m.tcos.n_terms,
            m.tsin.n_terms,
            m.tload.n_terms,
            m.fg.n_terms,
        )

        # We do the following (updating the basis functions manually) for speed
        # If we don't pass "init_basis" explicitly, the whole basis function has to
        # be recomputed, whereas if we just multiply the new/old K, we get a decent
        # speedup.
        K = np.vstack(extra_basis.values())
        new_extra_basis = np.repeat(K, repeater, axis=0)

        return attr.evolve(
            linear_model,
            init_basis=data["raw_basis"] * new_extra_basis,
            model=attr.evolve(linear_model.model, extra_basis=extra_basis),
        )

    @classmethod
    def from_labcal(
        cls,
        labcal,
        calobs,
        q_ant,
        qvar_ant,
        loads: dict | None = None,
        loss: float | np.ndarray = 1.0,
        loss_temp: float | np.ndarray = 0.0,
        fg_model=LinLog(n_terms=5),
        as_sim: tuple[str] = (),
        cal_noise="data",
        field_freq: np.ndarray = attr.NOTHING,
        bm_corr: float | np.ndarray = 1.0,
        s11_systematic_params=None,
        cterms=None,
        wterms=None,
        add_noise=True,
        seed=1234,
        **kwargs,
    ):
        if loads is None:
            loads = calobs.loads

        if as_sim == "all":
            as_sim = tuple(calobs.loads.keys())

        nwfg_model = NoiseWavesPlusFG.from_labcal(
            labcal,
            calobs,
            loads=loads,
            fg_model=fg_model,
            field_freq=field_freq,
            loss=loss,
            bm_corr=bm_corr,
            cterms=cterms,
            wterms=wterms,
        )

        freq = calobs.freq.freq.to_value("MHz")
        if s11_systematic_params:
            lna = S11Systematic(
                freq=freq,
                measured=calobs.receiver.s11_model(freq),
                params=s11_systematic_params.get("rcv", {}),
                name="rcv",
            )
            lna_ant = S11Systematic(
                freq=field_freq.to_value("MHz"),
                measured=calobs.receiver.s11_model(field_freq.to_value("MHz")),
                params=s11_systematic_params.get("rcv", {}),
                name="rcvant",
            )

            s11_systematics = tuple(
                S11Systematic(
                    freq=freq,
                    measured=load.reflections.s11_model(freq),
                    params=s11_systematic_params.get(src, {}),
                    name=src,
                )
                for src, load in loads.items()
            )

            ant_s11 = S11Systematic(
                freq=field_freq.to_value("MHz"),
                measured=labcal.antenna_s11_model(field_freq),
                params=s11_systematic_params.get("ant", {}),
                name="ant",
            )

            noise_wave_coeffs = (
                NoiseWaveCoefficients(
                    source_names=tuple(loads.keys()) + ("ant",),
                    components=s11_systematics + (lna, lna_ant, ant_s11),
                    as_dict=True,
                ),
            )
        else:
            noise_wave_coeffs = ()

        k0 = {
            src: rcf.get_K(
                gamma_ant=gamma_src(nwfg_model._freq(src)),
                gamma_rec=nwfg_model.gamma_rec(nwfg_model._freq(src)),
            )[0]
            for src, gamma_src in nwfg_model.gamma_src.items()
        }

        q = {
            name: simulate_q_from_calobs(calobs, name)
            if name in as_sim
            else load.spectrum.averaged_Q
            for name, load in loads.items()
        }

        q["ant"] = q_ant

        T = {
            name: load.temp_ave * np.ones(labcal.calobs.freq.n)
            for name, load in loads.items()
        }
        qvar = {"ant": qvar_ant}

        if cal_noise == "data" or isinstance(cal_noise, dict):
            qvar.update(
                {
                    name: load.spectrum.variance_Q / load.spectrum.n_integrations
                    for name, load in loads.items()
                }
            )
        else:
            qvar.update({name: cal_noise * np.ones(calobs.freq.n) for name in loads})

        if add_noise:
            np.random.seed(seed)
            for k in q:
                if k in as_sim:
                    if isinstance(cal_noise, dict):
                        q[k] += cal_noise[k]
                    else:
                        q[k] += np.random.normal(scale=np.sqrt(qvar[k]))

        raw_bases = nwfg_model.get_linear_model(with_k=False).basis

        data = {
            "q": q,
            "T": T,
            "k0": k0,
            "data_variance": qvar,
            "loss": loss,
            "bm_corr": bm_corr,
            "loss_temp": loss_temp,
            "gamma_src": {
                name: source.s11_model(freq) for name, source in loads.items()
            },
            "gamma_rcv": calobs.receiver.s11_model(freq),
            "raw_basis": raw_bases,
        }

        if not len(nwfg_model.field_freq) == len(q["ant"]) == len(qvar["ant"]):
            raise ValueError(
                "field_freq, q_ant and qvar_ant must be of the same shape."
            )

        return cls(
            nwfg_model=nwfg_model,
            data=data,
            s11_systematics=noise_wave_coeffs,
            **kwargs,
        )

    def get_linear_coefficients(
        self, params=None, ctx=None, fit=None, linear_params=None, src="ant"
    ) -> tuple[np.ndarray, np.ndarray]:
        """Get the linear coefficients that convert Q into a calibrated temperature.

        The equation is T_cal = a*Q + b

        This takes the input parameters (for TNS and T21), and computes the best-fit
        for the linear parameters.
        """
        if ctx is None:
            ctx = self.partial_linear_model.get_ctx(params=params)

        if fit is None and linear_params is None:
            fit = self.partial_linear_model.reduce_model(params=params)[0]
            params = fit.model_parameters
            model = fit.fit.model
            x = fit.fit.x
        elif fit is None:
            model = self.nwfg_model.linear_model.model
            x = self.nwfg_model.linear_model.x
            params = linear_params
        elif linear_params is None:
            model = fit.fit.model
            x = fit.fit.x
            params = fit.model_parameters
        else:
            x = fit.fit.x
            model = fit.fit.model
            params = linear_params

        idx = self.nwfg_model._get_idx(src)

        if src == "ant":
            a = ctx["tns_field"] / self.data["k0"]["ant"]
        else:
            a = ctx["tns"] / self.data["k0"][src]

        b = (
            -(
                model.get_model("tunc", x=x, with_extra=True, parameters=params)
                + model.get_model("tcos", x=x, with_extra=True, parameters=params)
                + model.get_model("tsin", x=x, with_extra=True, parameters=params)
                + model.get_model("tload", x=x, with_extra=True, parameters=params)
            )[idx]
            / self.data["k0"][src]
        )
        return a, b

    def recalibrated_sky_temp(
        self, params=None, ctx=None, fit=None, linear_params=None, a=None, b=None
    ) -> np.ndarray:
        """Get calibrated temperature of the data at a certain set of parameters.

        This takes the input parameters (for TNS and T21), computes the best-fit
        for the linear parameters, and applies the resulting calibration to the
        input field data.

        i.e. it gets (Tns*Q - K*Tnw)/K0
        """
        if a is None or b is None:
            a, b = self.get_linear_coefficients(
                params=params, ctx=ctx, fit=fit, linear_params=linear_params, src="ant"
            )

        return (a * self.data["q"]["ant"] + b - self.data["loss_temp"]) / (
            self.data["bm_corr"] * self.data["loss"]
        )

    def recalibrated_source_temp(self, src: str, params=None) -> np.ndarray:
        a, b = self.get_linear_coefficients(params=params, src=src)
        return a * self.data["q"][src] + b

    def sky_temp_model(self, params=None) -> tuple[np.ndarray, np.ndarray]:
        fit, data, var = self.partial_linear_model.reduce_model(params=params)
        freq = self.nwfg_model._freq("ant")

        fg = fit.fit.model.fg(x=freq)  # there's no K[0] or loss in this
        ctx = self.partial_linear_model.get_ctx(params=params)

        eor = ctx["eor_spectrum"]

        return fg, eor

    def cal_temp_model(self, src):
        return self.data["T"][src]

    def plot_calibrated_temps(
        self, params=None, fig=None, ax=None, labcal=None, fid_resid=None
    ):
        if labcal is not None:
            calobs = labcal.calobs

        if fig is None:
            fig, ax = plt.subplots(
                5,
                2,
                sharex=True,
                gridspec_kw={"wspace": 0.05, "hspace": 0.05},
                figsize=(5, 10),
            )

        for i, src in enumerate(self.src_names):
            if src == "ant":
                recal_temp = self.recalibrated_sky_temp(params=params)
                eor, fg = self.sky_temp_model(params=params)
                recal_temp_model = eor + fg
            else:
                recal_temp = self.recalibrated_source_temp(src=src, params=params)
                recal_temp_model = self.cal_temp_model(src)
            freq = self.nwfg_model._freq(src)

            ax[i, 0].plot(freq, recal_temp, label="Calibrated Data")
            ax[i, 0].plot(freq, recal_temp_model, label="Model", ls="--")
            ax[i, 0].set_ylabel(src)
            ax[i, 1].plot(freq, recal_temp - recal_temp_model)

            if labcal is not None and src != "ant":
                fid_data = calobs.calibrate(src, q=self.data["q"][src])
                ax[i, 0].plot(freq, fid_data, label="Fiducial Cal Data")
                ax[i, 1].plot(
                    freq,
                    fid_data - recal_temp_model,
                    label="Fid Resid",
                )

            elif labcal is not None and src == "ant":
                fid_data = (
                    labcal.calibrate_q(self.data["q"]["ant"], freq=freq)
                    - self.data["loss_temp"]
                ) / (self.data["bm_corr"] / self.data["loss"])
                ax[i, 0].plot(freq, fid_data)
                ax[i, 1].plot(freq, fid_data - recal_temp, label="Recal vs Labcal")
                if fid_resid is not None:
                    ax[i, 1].plot(freq, fid_resid)

        ax[0, 0].legend()
        ax[0, 1].legend()
        ax[-1, 1].legend()
        return fig, ax

    def plot_models(
        self, params=None, calobs=None, fig=None, ax=None, plot_zscore=True
    ):
        fit, data, var = self.partial_linear_model.reduce_model(params=params)

        if not plot_zscore:
            var = np.ones_like(var)

        if fig is None:
            fig, ax = plt.subplots(
                5,
                2,
                sharex=True,
                gridspec_kw={"wspace": 0.05, "hspace": 0.05},
                figsize=(5, 10),
            )

        n = 0
        mdata = fit.evaluate()
        for i, src in enumerate(self.src_names):
            freq = self.nwfg_model._freq(src)
            ax[i, 0].plot(freq, data[n : n + len(freq)], label="Recalibrated Data")
            ax[i, 0].plot(
                freq, mdata[n : n + len(freq)], label="best-Fit Linear Model", ls="--"
            )
            ax[i, 0].set_ylabel(src)
            ax[i, 1].plot(
                freq,
                (data[n : n + len(freq)] - mdata[n : n + len(freq)])
                / np.sqrt(var[n : n + len(freq)]),
                label="Recal Resid",
            )

            if calobs is not None and src != "ant":
                fid_data = (
                    calobs.C1() * calobs.t_load_ns * self.data["q"][src]
                    - self.data["k0"][src] * self.data["T"][src]
                )
                ax[i, 0].plot(calobs.freq.freq, fid_data, label="Fiducial Cal Data")
                fid_model_q = (
                    calobs.decalibrate(temp=self.data["T"][src], load=src)
                    - calobs.t_load
                ) / calobs.t_load_ns
                fid_model = (
                    calobs.C1() * calobs.t_load_ns * fid_model_q
                    - self.data["k0"][src] * self.data["T"][src]
                )
                if plot_zscore:
                    fid_var = (calobs.C1() * calobs.t_load_ns) ** 2 * self.data[
                        "data_variance"
                    ][src]
                else:
                    fid_var = 1

                ax[i, 0].plot(calobs.freq.freq, fid_model, label="Fiducial Cal Model")

                ax[i, 1].plot(
                    calobs.freq.freq,
                    (fid_data - fid_model) / np.sqrt(fid_var),
                    label="Fid Resid",
                )

            n += len(freq)

        ax[0, 0].legend()
        ax[0, 1].legend()
        return fig, ax


@attr.s(frozen=True, kw_only=True)
class LinearFG:
    """Classic traditional EoR fit to a calibrated sky spectrum.

    This defines a ``partial_linear_model`` that simply fits the EoR + FG, where the FG
    is assumed to be a linear model. While this is not the ultimate in flexibility,
    since the FG aren't always linear, it is fast.
    """

    freq: np.ndarray = attr.ib()
    t_sky: np.ndarray = attr.ib()
    var: np.ndarray = attr.ib()
    fg: Model = attr.ib()
    eor: AbsorptionProfile = attr.ib()

    @eor.default
    def _eorcmp(self):
        return (AbsorptionProfile(freqs=self.freq, params=("A", "w", "tau", "nu0")),)

    @cached_property
    def partial_linear_model(self):
        return PartialLinearModel(
            linear_model=self.fg.at(x=self.freq),
            data={"t_sky": self.t_sky, "data_variance": self.var},
            components=(self.eor,),
            data_func=self.transform_data,
            variance_func=None,
        )

    def transform_data(self, ctx: dict, data: dict):
        return data["t_sky"] - ctx["eor_spectrum"]



"""
Components for performing calibration on raw data.
"""
import attr
import logging
import numpy as np
from attr import validators as vld
from cached_property import cached_property
from edges_cal import receiver_calibration_func as rcf
from edges_cal.cal_coefficients import CalibrationObservation
from edges_cal.receiver_calibration_func import power_ratio
from edges_cal.s11 import LoadS11, Receiver
from edges_io.logging import logger
from yabf import Component, Parameter, ParameterVector


def _log_level_converter(val):
    if isinstance(val, int):
        return val
    elif isinstance(val, str):
        try:
            return getattr(logging, val.upper())
        except AttributeError:
            raise ValueError(f"{val} is not an available logging level")
    else:
        raise TypeError("log_level must be int or str")


@attr.s(frozen=True, cache_hash=True)
class _CalibrationQ(Component):
    """Base Component providing calibration Q_P."""

    path = attr.ib(kw_only=True, default="", validator=vld.instance_of(str))
    calobs_args = attr.ib(
        kw_only=True, default={}, converter=dict, validator=vld.instance_of(dict)
    )
    _log_level = attr.ib(
        kw_only=True, default=logging.WARNING, converter=_log_level_converter
    )
    _calobs = attr.ib(
        kw_only=True,
        default=None,
        validator=vld.optional(vld.instance_of(CalibrationObservation)),
    )

    @cached_property
    def calobs(self):
        if self._calobs is not None:
            return self._calobs

        if not self.path:
            raise ValueError("if calobs is not given, path must be")

        logger.setLevel(self._log_level)
        return CalibrationObservation(path=self.path, **self.calobs_args)

    @cached_property
    def base_parameters(self):
        c1_terms = ParameterVector(
            "C1", fiducial=0, length=self.calobs.cterms, latex=r"C^1_%s"
        ).get_params()
        c2_terms = ParameterVector(
            "C2", fiducial=0, length=self.calobs.cterms, latex=r"C^2_%s"
        ).get_params()

        tunc_terms = ParameterVector(
            "Tunc", fiducial=0, length=self.calobs.wterms, latex=r"T^{\rm unc}_{%s}"
        ).get_params()
        tcos_terms = ParameterVector(
            "Tcos", fiducial=0, length=self.calobs.wterms, latex=r"T^{\rm cos}_{%s}"
        ).get_params()
        tsin_terms = ParameterVector(
            "Tsin", fiducial=0, length=self.calobs.wterms, latex=r"T^{\rm sin}_{%s}"
        ).get_params()

        return tuple(c1_terms + c2_terms + tunc_terms + tcos_terms + tsin_terms)

    @cached_property
    def freq(self):
        return self.calobs.freq.freq

    @cached_property
    def data_mask(self):
        """The data itself is averaged_Q from the LoadSpectrum, which may involve different
        frequencies than the calibration itself. Here we get which elements to actually use."""
        mask = []
        for i, flag in enumerate(self.calobs.open.spectrum.freq.mask):
            if not flag:
                continue
            else:
                mask.append(self.calobs.freq.mask[i])
        return np.array(mask, dtype=bool)

    @cached_property
    def freq_recentred(self):
        return np.linspace(-1, 1, len(self.freq))

    @cached_property
    def provides(self):
        return [f"{self.name}_calibration_q", "data_mask", "cal_curves", "freq_obj"]

    def get_calibration_curves(self, params):
        # Put coefficients in backwards, because that's how the polynomial works.
        c1, c2, tu, tc, ts = self.get_cal_funcs(params)

        return (
            c1(self.freq_recentred),
            c2(self.freq_recentred),
            tu(self.freq_recentred),
            tc(self.freq_recentred),
            ts(self.freq_recentred),
        )

    def get_cal_funcs(self, params=None):
        params = self._fill_params(params)

        # Put coefficients in backwards, because that's how the polynomial works.
        c1_poly = np.poly1d(
            [params[f"C1_{i}"] for i in range(self.calobs.cterms)[::-1]]
        )
        c2_poly = np.poly1d(
            [params[f"C2_{i}"] for i in range(self.calobs.cterms)[::-1]]
        )
        tunc_poly = np.poly1d(
            [params[f"Tunc_{i}"] for i in range(self.calobs.wterms)[::-1]]
        )
        tcos_poly = np.poly1d(
            [params[f"Tcos_{i}"] for i in range(self.calobs.wterms)[::-1]]
        )
        tsin_poly = np.poly1d(
            [params[f"Tsin_{i}"] for i in range(self.calobs.wterms)[::-1]]
        )

        return c1_poly, c2_poly, tunc_poly, tcos_poly, tsin_poly


@attr.s(frozen=True)
class CalibratorQ(_CalibrationQ):
    """Component providing calibration Q_P for calibrator sources ambient, hot_load,
    open, short.

    Parameters
    ----------
    """

    @cached_property
    def s11_models(self):
        return {
            "open": self.calobs.open.s11_model(self.freq),
            "short": self.calobs.short.s11_model(self.freq),
            "hot_load": self.calobs.hot_load.s11_model(self.freq),
            "ambient": self.calobs.ambient.s11_model(self.freq),
            "lna": self.calobs.receiver.s11_model(self.freq),
        }

    @cached_property
    def Ks(self):
        return {
            name: rcf.get_K(self.s11_models["lna"], self.s11_models[name])
            for name in self.s11_models
            if name != "lna"
        }

    def calculate(self, ctx=None, **params):
        scale, offset, tu, tc, ts = self.get_cal_funcs(params)

        Qp = {}
        for name, source in self.calobs._loads.items():
            temp_ant = source.spectrum.temp_ave

            a, b = rcf.get_linear_coefficients_from_K(
                self.Ks[name],
                scale(self.freq_recentred),
                offset(self.freq_recentred),
                tu(self.freq_recentred),
                tc(self.freq_recentred),
                ts(self.freq_recentred),
                t_load=300,
            )

            Qp[name] = ((temp_ant - b) / a - 300) / 400

        return (
            Qp,
            self.data_mask,
            {"c1": scale, "c2": offset, "tu": tu, "tc": tc, "ts": ts},
            self.calobs.freq,
        )


@attr.s(frozen=True)
class AntennaQ(_CalibrationQ):
    """Component providing calibration Q_P for calibrator sources ambient, hot_load,
    open, short.

    Parameters
    ----------
    antenna : :class:`~edges_cal.cal_coefficients.LoadS11` or
        :class:`~edges_cal.cal_coefficients.LoadSpectrum`
        The properties of the antenna. If a `LoadSpectrum`, assumes that the true temperature
        is known. If a `LoadS11`, assumes that the true temperature is forward-modelled
        by subcomponents.
    receiver : :class:`~edges_cal.cal_coefficients.LNA`
        The S11 of the reciever/LNA.
    """

    antenna = attr.ib(kw_only=True, validator=attr.validators.instance_of(LoadS11))
    receiver = attr.ib(kw_only=True, validator=attr.validators.instance_of(Receiver))

    @cached_property
    def freq(self):
        return self.antenna.freq.freq

    def calculate(self, ctx=None, **params):
        scale, offset, tu, tc, ts = self.get_calibration_curves(params)

        temp_ant = sum(v for k, v in ctx.items() if k.endswith("spectrum"))
        gamma_ant = self.antenna.get_s11_correction_model()(self.freq)

        return power_ratio(
            scale=scale,
            offset=offset,
            temp_cos=tc,
            temp_sin=ts,
            temp_unc=tu,
            temp_ant=temp_ant,
            gamma_ant=gamma_ant,
            gamma_rec=self.receiver.get_s11_correction_model()(self.freq),
            temp_noise_source=400,
            temp_load=300,
        )



import numpy as np
from astropy import units as u
from edges_cal.modelling import Polynomial, UnitTransform
from edges_cal.simulate import simulate_qant_from_calobs
from scipy import stats
from yabf import ParamVec


def get_tns_model(calobs, ideal=False):
    if ideal:
        p = np.array([1575, -175, 70.0, -17.5, 7.0, -3.5])
    else:
        p = calobs.C1_poly.coeffs[::-1] * calobs.t_load_ns

    t_ns_model = Polynomial(
        parameters=p,
        transform=UnitTransform(range=(calobs.freq.min.value, calobs.freq.max.value)),
    )

    t_ns_params = ParamVec(
        "t_lns",
        length=len(p),
        min=p - 100,
        max=p + 100,
        ref=[stats.norm(v, scale=1.0) for v in p],
        fiducial=p,
    )
    return t_ns_model, t_ns_params


def sim_antenna_q(labcal, calobs, fg, eor, ideal_tns=True, loss=1, bm_corr=1):
    spec = fg(x=eor.freqs) + eor()["eor_spectrum"]

    tns_model, _ = get_tns_model(calobs, ideal=ideal_tns)
    _scale_model = tns_model.with_params(
        np.array(tns_model.parameters) / calobs.t_load_ns
    )

    return simulate_qant_from_calobs(
        calobs,
        ant_s11=labcal.antenna_s11_model(eor.freqs),
        ant_temp=spec,
        scale_model=lambda f: _scale_model(f.to_value("MHz")),
        loss=loss,
        freq=eor.freqs * u.MHz,
        bm_corr=bm_corr,
    )
